import os
import subprocess

import mock
import pytest

import batou.utils
from batou.repository import MercurialRepository


def test_repository_hg_norepo(tmpdir):
    tmpdir = str(tmpdir)
    environment = mock.Mock()
    environment.root = tmpdir
    environment.base_dir = tmpdir
    os.chdir(tmpdir)

    with pytest.raises(batou.utils.CmdExecutionError):
        MercurialRepository(environment)

    subprocess.check_call(["hg", "init"])
    MercurialRepository(environment)


def test_repository_hg_show_upstream(tmpdir):
    tmpdir = str(tmpdir)
    environment = mock.Mock()
    environment.root = tmpdir
    environment.base_dir = tmpdir
    os.chdir(tmpdir)
    subprocess.check_call(["hg", "init"])

    repository = MercurialRepository(environment)
    environment.repository_url = "asdf"
    assert repository.upstream == "asdf"
    repository._upstream = None

    environment.repository_url = None
    with pytest.raises(batou.utils.CmdExecutionError):
        repository.upstream

    with open(".hg/hgrc", "w") as f:
        f.write(
            """\
[paths]
foobar = 1234
"""
        )

    with pytest.raises(AssertionError):
        repository.upstream

    with open(".hg/hgrc", "w") as f:
        f.write(
            """\
[paths]
foobar = 1234
default = ssh://test@example.com/repos
"""
        )

    assert repository.upstream == "ssh://test@example.com/repos"
    # Trigger cached access
    assert repository.upstream == "ssh://test@example.com/repos"


def test_repository_hg_verify(tmpdir):
    tmpdir = str(tmpdir)
    environment = mock.Mock()
    environment.root = tmpdir
    environment.base_dir = tmpdir
    environment.deployment.dirty = False

    os.chdir(tmpdir)

    subprocess.check_call(["hg", "init"])

    subprocess.check_call(["hg", "init", "remote"])

    # Create a second repo as target
    with open(".hg/hgrc", "w") as f:
        f.write(
            """\
[paths]
default = file:///{}/remote
""".format(
                tmpdir
            )
        )

    repository = MercurialRepository(environment)
    # Clean repositories are fine
    repository.verify()

    with open("asdf", "w") as f:
        f.write("foobar!")

    with pytest.raises(batou.utils.DeploymentError) as e:
        repository.verify()
    assert e.value.args[0] == "Uncommitted changes"

    subprocess.check_call(["hg", "commit", "-A", "-m", "test"])

    with pytest.raises(batou.utils.DeploymentError) as e:
        repository.verify()
    assert e.value.args[0] == "Outgoing changes"

    # We actually could deploy if we allow dirty deployments
    environment.deployment.dirty = True
    repository.verify()
    environment.deployment.dirty = False

    subprocess.check_call(["hg", "push"])
    repository.verify()


def test_repository_hg_ship_verify_after_ship(tmpdir):
    tmpdir = str(tmpdir)
    environment = mock.Mock()
    environment.root = tmpdir
    environment.base_dir = tmpdir
    environment.deployment.dirty = False

    os.chdir(tmpdir)
    subprocess.check_call(["hg", "init"])

    repository = MercurialRepository(environment)

    # We just want to ensure that the shipping validation code is fine
    def no_ship(host):
        pass

    repository._ship = no_ship

    host = mock.Mock()
    host.rpc.hg_update_working_copy.return_value = (
        "0000000000000000000000000000000000000000"
    )
    repository.update(host)

    with open("asdf", "w") as f:
        f.write("foobar")
    subprocess.check_call(["hg", "add", "asdf"])
    with pytest.raises(batou.RepositoryDifferentError):
        repository.update(host)

    environment.deployment.dirty = True
    repository.update(host)



import os
import os.path
import shlex
import socket
import tempfile
import textwrap
import uuid

import requests

import batou.utils
from batou import output
from batou.utils import cmd


class Provisioner(object):
    rebuild = False

    def __init__(self, name):
        self.name = name

    @classmethod
    def from_config_section(cls, section):
        return NotImplemented

    def provision(self, host):
        pass

    def suggest_name(self, name):
        return name


class FCDevProvisioner(Provisioner):
    target_host = None
    aliases = ()

    def _prepare_ssh(self, host):
        # XXX application / user-specific files
        # https://unix.stackexchange.com/questions/312988/understanding-home-configuration-file-locations-config-and-local-sha
        KNOWN_HOSTS_FILE = os.path.expanduser("~/.batou/known_hosts")

        if not os.path.exists(KNOWN_HOSTS_FILE):
            prefix_dir = os.path.dirname(KNOWN_HOSTS_FILE)
            if not os.path.exists(prefix_dir):
                os.makedirs(prefix_dir)
            with open(KNOWN_HOSTS_FILE, "w"):
                pass

        if self.rebuild:
            cmd(["ssh-keygen", "-R", host.name, "-f", KNOWN_HOSTS_FILE])

        ssh_config = []
        # We need to provide a version of the key that doesn't trigger
        # OpenSSHs "wrong permission" check.
        packaged_insecure_key = os.path.join(
            os.path.dirname(__file__), "insecure-private.key"
        )
        local_insecure_key = os.path.abspath("insecure-private.key")
        with open(local_insecure_key, "w") as f_target:
            with open(packaged_insecure_key) as f_packaged:
                f_target.write(f_packaged.read())
        os.chmod(local_insecure_key, 0o600)

        ssh_config.append(
            """
Host {hostname} {aliases}
    HostName {hostname}
    ProxyJump {target_host}
    User developer
    IdentityFile {insecure_private_key}
    StrictHostKeyChecking no
    UserKnownHostsFile {known_hosts}
""".format(
                hostname=host.name,
                aliases=" ".join(host._aliases),
                target_host=self.target_host,
                known_hosts=KNOWN_HOSTS_FILE,
                insecure_private_key=local_insecure_key,
            )
        )

        # More generic includes need to go last because parameters are
        # set by a match-first and things like User settings for * may
        # otherwise collide with our very specific settings. See
        # `man 5 ssh_config`.
        if os.path.exists(os.path.expanduser("~/.ssh/config")):
            ssh_config.append("Host *\n  Include ~/.ssh/config")
        if os.path.exists("ssh_config"):
            ssh_config.append(
                "Host *\n  Include {}".format(os.path.abspath("ssh_config"))
            )

        # Place this in the deployment base directory persistently and
        # keep updating it. This helps users to also interact with containers
        # by running `ssh -F ssh_config_dev mycontainer`
        self.ssh_config_file = os.path.abspath(
            "ssh_config_{}".format(host.environment.name)
        )
        with open(self.ssh_config_file, "w") as f:
            f.write("\n".join(ssh_config))

    def configure_host(self, host, config):
        # Extract provisioning-specific config from host

        # Establish host-specific list of aliases and their public FQDNs.
        host.aliases.update(
            {
                alias.strip(): f"{alias}.{host.name}.{self.target_host}"
                for alias in config.get("provision-aliases", "").strip().split()
            }
        )

        # Development containers have a different internal address for the
        # external aliases so that we need to explicitly override the resolver
        # so that services like nginx can bind to the correct local address.
        try:
            addr = batou.utils.resolve(host.name)
            for alias_fqdn in host.aliases.values():
                output.annotate(
                    f" alias override v4 {alias_fqdn} -> {addr}", debug=True
                )
                batou.utils.resolve_override[alias_fqdn] = addr
        except (OSError, ValueError):
            # We used to catch socket.gaierror here, but reading the
            # docs correctly it can raise any OSError (and socket and gaierror
            # are subclasses)
            pass

        try:
            addr = batou.utils.resolve_v6(host.name)
            for alias_fqdn in host.aliases.values():
                output.annotate(
                    f" alias override v6 {alias_fqdn} -> {addr}", debug=True
                )
                batou.utils.resolve_v6_override[alias_fqdn] = addr
        except (OSError, ValueError):
            # We used to catch socket.gaierror here, but reading the
            # docs correctly it can raise any OSError (and socket and gaierror
            # are subclasses)
            pass

    def summarize(self, host):
        for alias, fqdn in host.aliases.items():
            output.line(f" 🌐 https://{fqdn}/")

    def _initial_provision_env(self, host):
        return NotImplemented

    def provision(self, host):
        self._prepare_ssh(host)

        rsync_path = ""
        if host.environment.service_user:
            rsync_path = (
                f'--rsync-path="sudo -u {host.environment.service_user} '
                f'rsync"'
            )
        env = self._initial_provision_env(host)
        env["SSH_CONFIG"] = self.ssh_config_file
        env["RSYNC_RSH"] = "ssh -F {}".format(self.ssh_config_file)
        if self.rebuild:
            env["PROVISION_REBUILD"] = "1"
        # Add all component variables (uppercased) to the environment
        # $COMPONENT_<component>_<variable>
        # this includes environment overrides and secrets.
        for root_name, root in host.components.items():
            root = root.factory
            for name in dir(root):
                if name.startswith("_"):
                    continue
                if name in ["changed", "workdir", "namevar"]:
                    continue
                value = getattr(root, name)
                if callable(value):
                    continue
                if isinstance(value, property):
                    continue
                key = f"COMPONENT_{root_name}_{name}"
                key = key.upper()
                env[key] = str(value)
            for name, value in host.environment.overrides.get(
                root_name, {}
            ).items():
                key = f"COMPONENT_{root_name}_{name}"
                key = key.upper()
                env[key] = str(value)

        seed_script = ""
        seed_basedir = f"environments/{host.environment.name}"
        seed_script_file = f"{seed_basedir}/provision.sh"
        if os.path.exists(seed_script_file):
            output.annotate(f"    Including {seed_script_file}")
            seed_raw_script = open(seed_script_file).read()
            seed_script += textwrap.dedent(
                f"""\
                # BEGIN CUSTOM SEED SCRIPT
                (
                    cd {seed_basedir}
                    {seed_raw_script}
                )
                # END CUSTOM SEED SCRIPT
                """
            )

        seed_nixos_file = f"environments/{host.environment.name}/provision.nix"
        if (
            os.path.exists(seed_nixos_file)
            and "provision.nix" not in seed_script
        ):
            output.annotate(f"    Including {seed_nixos_file}")
            seed_script = (
                textwrap.dedent(
                    f"""\
                # BEGIN AUTOMATICALLY INCLUDED provision.nix
                (
                    cd {seed_basedir}
                    COPY provision.nix /etc/local/nixos/provision-container.nix
                )
                # END AUTOMATICALLY INCLUDED provision.nix
                """
                )
                + seed_script
            )

        seed_script = seed_script.strip()
        if not seed_script:
            output.annotate(
                f"No provisioning code found in "
                f"environments/{host.environment.name}/provision.nix or "
                f"environments/{host.environment.name}/provision.sh. "
                f"This might be unintentional.",
                yellow=True,
            )

        stdout = stderr = ""
        with tempfile.NamedTemporaryFile(
            mode="w+", prefix="batou-provision", delete=False
        ) as f:
            try:
                os.chmod(f.name, 0o700)
                # We're placing the ENV vars directly in the script because
                # that helps debugging a lot. We need to be careful to delete
                # it later, though, because it is likely to contain secrets.
                f.write(
                    self.SEED_TEMPLATE.format(
                        seed_script=seed_script,
                        rsync_path=rsync_path,
                        ENV=batou.utils.export_environment_variables(env),
                    )
                )
                f.close()
                stdout, stderr = cmd(f.name)
            except Exception:
                raise
            else:
                if "__FC_MANAGE_DEFECT_INDICATOR__" in stdout:
                    stdout = stdout.replace(
                        "__FC_MANAGE_DEFECT_INDICATOR__", ""
                    )
                    output.section(
                        "Errors detected during provisioning", red=True
                    )
                    output.line("STDOUT")
                    output.annotate(stdout)
                    output.line("STDERR")
                    output.annotate(stderr)
                    output.line(
                        "WARNING: Continuing deployment optimistically "
                        "despite provisioning errors. Check errors above this "
                        "line first if encountering subsequent errors.",
                        yellow=True,
                    )
                else:
                    output.line("STDOUT", debug=True)
                    output.annotate(stdout, debug=True)
                    output.line("STDERR", debug=True)
                    output.annotate(stderr, debug=True)
            finally:
                # The script includes secrets so we must be sure that we delete
                # it.
                if output.enable_debug:
                    output.annotate(
                        (
                            f"Not deleting provision script "
                            f"{f.name} in debug mode!"
                        ),
                        red=True,
                    )
                    os.unlink(f.name)


class FCDevContainer(FCDevProvisioner):
    SEED_TEMPLATE = """\
#/bin/sh
set -e

{ENV}

ECHO() {{
    what=${{1?what to echo}}
    where=${{2?where to echo}}
    RUN "echo $what > $where"
}}

RUN() {{
    cmd=$@
    ssh -F $SSH_CONFIG $PROVISION_CONTAINER "$cmd"
}}

COPY() {{
    what=${{1?what to copy}}
    where=${{2?where to copy}}
    rsync -avz --no-l --safe-links {rsync_path} $what $PROVISION_CONTAINER:$where
}}

if [ ${{PROVISION_REBUILD+x}} ]; then
    ssh $PROVISION_HOST sudo fc-build-dev-container destroy $PROVISION_CONTAINER
fi

ssh $PROVISION_HOST sudo fc-build-dev-container ensure $PROVISION_CONTAINER $PROVISION_CHANNEL "'$PROVISION_ALIASES'"

{seed_script}

# We experimented with hiding errors in this fc-manage run to allow
# partially defective NixOS configurations to be repaired with subsequent
# deployment actions, so we do have to continue here.
# However, we need to show if something goes wrong so that users have an
# indication that the cause might be here. Especially if provisioning
# is half-baked. Unfortunately we cant' decide whether the error is caused
# by the provisioning or the deployment step.

set +e
RUN sudo -i fc-manage -c
result=$?
if [ "$result" -ne "0" ]; then
    echo "__FC_MANAGE_DEFECT_INDICATOR__"
fi

"""  # noqa: E501 line too long
    target_host = None
    channel = None
    aliases = ()

    @classmethod
    def from_config_section(cls, name, section):
        instance = FCDevContainer(name)
        instance.target_host = section["host"]
        instance.channel = section["channel"]
        return instance

    def suggest_name(self, name):
        config = "-F ssh_config" if os.path.exists("ssh_config") else ""
        out, _ = cmd(
            "ssh {config} {target_host} 'nixos-container list'".format(
                config=config, target_host=self.target_host
            )
        )
        names = filter(None, [x.strip() for x in out.splitlines()])
        while True:
            name = uuid.uuid4().hex[:11]
            if name not in names:
                return name

    def _initial_provision_env(self, host):
        return {
            "PROVISION_CONTAINER": host.name,
            "PROVISION_HOST": self.target_host,
            "PROVISION_CHANNEL": self.channel,
            "PROVISION_ALIASES": " ".join(host.aliases.keys()),
        }


class FCDevVM(FCDevProvisioner):
    SEED_TEMPLATE = """\
#/bin/sh
set -e

{ENV}

ECHO() {{
    what=${{1?what to echo}}
    where=${{2?where to echo}}
    RUN "echo $what > $where"
}}

RUN() {{
    cmd=$@
    ssh -F $SSH_CONFIG $PROVISION_VM "$cmd"
}}

COPY() {{
    what=${{1?what to copy}}
    where=${{2?where to copy}}
    rsync -avz --no-l --safe-links {rsync_path} $what $PROVISION_VM:$where
}}

if [ ${{PROVISION_REBUILD+x}} ]; then
    ssh $PROVISION_HOST sudo fc-devhost destroy $PROVISION_VM
fi

cli_args="--memory $PROVISION_VM_MEMORY\\
  --cpu $PROVISION_VM_CORES \\
  --aliases \\"'$PROVISION_ALIASES'\\""

# Error handling is done in the python part
if [ -n "$PROVISION_HYDRA_EVAL" ]; then
    cli_args="${{cli_args}} --hydra-eval $PROVISION_HYDRA_EVAL"
else
    cli_args="${{cli_args}} --image-url $PROVISION_IMAGE --channel-url $PROVISION_CHANNEL"
fi

ssh $PROVISION_HOST sudo fc-devhost ensure \\
    $cli_args \\
    $PROVISION_VM

{seed_script}

# We experimented with hiding errors in this fc-manage run to allow
# partially defective NixOS configurations to be repaired with subsequent
# deployment actions, so we do have to continue here.
# However, we need to show if something goes wrong so that users have an
# indication that the cause might be here. Especially if provisioning
# is half-baked. Unfortunately we cant' decide whether the error is caused
# by the provisioning or the deployment step.

set +e
RUN sudo -i fc-manage -c
result=$?
if [ "$result" -ne "0" ]; then
    echo "__FC_MANAGE_DEFECT_INDICATOR__"
fi

"""  # noqa: E501 line too long
    target_host = None
    aliases = ()
    memory = None
    cores = None

    hydra_eval = ""  # deprecated
    channel_url = ""
    image_url = ""

    @classmethod
    def from_config_section(cls, name, section):
        instance = FCDevVM(name)
        instance.target_host = section["host"]
        instance.memory = section["memory"]
        instance.cores = section["cores"]

        if "release" in section:
            resp = requests.get(section["release"])
            resp.raise_for_status()
            release_info = resp.json()
            instance.channel_url = release_info["channel_url"]
            instance.image_url = release_info["devhost_image_url"]
        elif "hydra-eval" in section:
            instance.hydra_eval = section["hydra-eval"]
        else:
            raise ValueError(
                "Either `release` or `hydra-eval` must be set in the provisioner config section."
            )
        return instance

    def suggest_name(self, name):
        config = "-F ssh_config" if os.path.exists("ssh_config") else ""
        out, _ = cmd(f"ssh {config} {self.target_host} 'sudo fc-devhost list'")
        names = filter(None, [x.strip() for x in out.splitlines()])
        while True:
            name = uuid.uuid4().hex[:8]
            if name not in names and not name.isnumeric():
                return name

    def _initial_provision_env(self, host):
        return {
            "PROVISION_VM": host.name,
            "PROVISION_HOST": self.target_host,
            "PROVISION_ALIASES": " ".join(host.aliases.keys()),
            "PROVISION_HYDRA_EVAL": self.hydra_eval,
            "PROVISION_CHANNEL": self.channel_url,
            "PROVISION_IMAGE": self.image_url,
            "PROVISION_VM_MEMORY": self.memory,
            "PROVISION_VM_CORES": self.cores,
        }



import os
import os.path
import time

from batou import UpdateNeeded, output
from batou.component import Attribute, Component, ConfigString, handle_event
from batou.lib.buildout import Buildout
from batou.lib.file import Directory, File
from batou.lib.logrotate import RotatedLogfile
from batou.lib.nagios import ServiceCheck
from batou.lib.service import Service
from batou.utils import Address, CmdExecutionError


class Program(Component):

    program_section = """
[program:{{component.name}}]
command = {{component.command}} {{component.args}}
process_name = {{component.name}}
directory = {{component.directory}}
priority = {{component.priority}}
stderr_logfile = {{component.supervisor.logdir.path}}/{{component.name}}.log
stderr_logfile_backups = 0
stderr_logfile_maxbytes = 0
stdout_logfile = {{component.supervisor.logdir.path}}/{{component.name}}.log
stdout_logfile_backups = 0
stdout_logfile_maxbytes = 0
redirect_stderr = true
{% for k, v in component.options.items() %}
{{k}} = {{v}}
{% endfor -%}
"""

    namevar = "name"

    # Can this be deployed while running or should the process be stopped
    # prior to deploying it?
    # hot = it can keep running
    # cold = needs to be stopped
    deployment = "hot"

    command = None
    command_absolute = True
    options = {}
    args = ""
    priority = 10
    directory = None
    dependencies = None
    enable = True

    def configure(self):
        self.supervisor = self.require_one("supervisor", self.host)
        if self.dependencies is None:
            self.dependencies = (self.parent,)
        if not self.command:
            raise ValueError(
                "`command` option missing for program {}".format(self.name)
            )
        if not self.directory:
            self.directory = self.workdir
        if self.command_absolute:
            self.command = os.path.normpath(
                os.path.join(self.workdir, self.command)
            )

        if "startsecs" not in self.options:
            self.options["startsecs"] = 5
        if "startretries" not in self.options:
            self.options["startretries"] = 3
        self.supervisor.max_startup_delay = max(
            int(self.options["startsecs"]), self.supervisor.max_startup_delay
        )

        self.config = self.expand(
            "{{component.supervisor.program_config_dir.path}}/"
            "{{component.name}}.conf"
        )
        self += File(self.config, content=self.expand(self.program_section))

    def ctl(self, args, **kw):
        command = "{}/bin/supervisorctl".format(self.supervisor.workdir)
        return self.cmd("{} {}".format(command, args), **kw)

    def verify(self):
        if not self.supervisor.enable:
            return
        for dependency in self.dependencies:
            dependency.assert_no_changes()
        running = self.is_running()
        if (self.enable and not running) or (not self.enable and running):
            raise UpdateNeeded()

    def update(self):
        self.ctl("reread")
        self.ctl("update")
        if self.enable:
            self.ctl("restart {}".format(self.name))
        else:
            if self.is_running():
                self.ctl("stop {}".format(self.name))
            return
        if self.supervisor.wait_for_running:
            for retry in range(self.options["startsecs"]):
                time.sleep(1)
                if self.is_running():
                    return
            raise RuntimeError(
                'Program "{}" did not start up'.format(self.name)
            )

    def is_running(self):
        out, err = self.ctl(
            "status {}".format(self.name), ignore_returncode=True
        )
        return "RUNNING" in out

    # Keep track whether
    _evaded = False

    @handle_event("before-update", "precursor")
    def evade(self, component):
        if self.deployment == "hot":
            return
        if self._evaded:
            return
        # Only try once. Keep going anyway.
        self._evaded = True
        output.annotate(
            "\u2623 Stopping {} for cold deployment".format(self.name)
        )
        try:
            self.ctl("stop {}".format(self.name))
        except Exception:
            pass


class Eventlistener(Program):

    program_section = """
[eventlistener:{{component.name}}]
command = {{component.command}} {{component.args}}
events = {{component.events}}
process_name={{component.name}}
"""

    namevar = "name"

    events = ("TICK_60",)
    command = None
    args = None

    def configure(self):
        if not isinstance(self.events, str):
            self.events = ",".join(self.events)
            # Not sure what's right. We only use eventlisteners with superlance
            # which lives in the supervisor's workdir. However, the
            # EventListener component gets instanciated as a sub-component of
            # the actual component using it - which doesn't know about the path
            # to the superlance plugins. :/
            self.supervisor = self.require_one("supervisor", self.host)
            self.command = os.path.normpath(
                os.path.join(self.supervisor.workdir, self.command)
            )

        super(Eventlistener, self).configure()


class Supervisor(Component):

    address = Attribute(Address, default=ConfigString("localhost:9001"))
    buildout_version = Attribute(default="3.0.1")
    setuptools_version = Attribute(default="68.0.0")
    wheel_version = Attribute(default="0.40.0")
    buildout_cfg = os.path.join(
        os.path.dirname(__file__), "resources", "supervisor.buildout.cfg"
    )
    supervisor_conf = os.path.join(
        os.path.dirname(__file__), "resources", "supervisor.conf"
    )

    program_config_dir = None
    logdir = None
    loglevel = "info"

    logrotate = Attribute("literal", default=False)
    nagios = Attribute("literal", default=False)

    # Allows turning "everything off" via environment configuration
    enable = Attribute("literal", default=True)
    # Hot deployments: if supervisor is already running stuff - keep them
    # running
    # Cold deployments: if supervisor is already running: let it run
    # but shutdown all processes before continuing
    deployment_mode = Attribute(str, default="hot")
    max_startup_delay = Attribute(int, default=0)
    wait_for_running = Attribute("literal", default=True)
    pidfile = Attribute(str, default=ConfigString("supervisord.pid"), map=True)
    socketpath = Attribute(
        str,
        default=ConfigString("{{component.workdir}}/supervisor.sock"),
        map=True,
    )
    check_contact_groups = None

    def configure(self):
        self.provide("supervisor", self)

        buildout_cfg = File("buildout.cfg", source=self.buildout_cfg)
        self += Buildout(
            version=self.buildout_version,
            setuptools=self.setuptools_version,
            config=buildout_cfg,
            python="3",
            wheel=self.wheel_version,
        )

        self.program_config_dir = Directory("etc/supervisor.d", leading=True)
        self += self.program_config_dir
        self += File("etc/supervisord.conf", source=self.supervisor_conf)
        self.logdir = Directory("var/log", leading=True)
        self += self.logdir

        postrotate = self.expand(
            "kill -USR2 $({{component.workdir}}/bin/supervisorctl pid)"
        )

        if self.logrotate:
            self += RotatedLogfile("var/log/*.log", postrotate=postrotate)

        self += Service("bin/supervisord", pidfile=self.pidfile)
        service = self._

        if self.enable:
            self += RunningSupervisor(service)
        else:
            self += StoppedSupervisor()

        self += File(
            "check_supervisor",
            mode=0o755,
            source=os.path.join(
                os.path.dirname(__file__), "resources", "check_supervisor.py.in"
            ),
        )

        if self.nagios:
            self += ServiceCheck(
                "Supervisor programs",
                nrpe=True,
                contact_groups=self.check_contact_groups,
                command=self.expand("{{component.workdir}}/check_supervisor"),
            )


class RunningHelper(Component):
    def is_running(self):
        try:
            pid, err = self.cmd("bin/supervisorctl pid")
            try:
                int(pid) > 0
            except ValueError:
                return False
        except CmdExecutionError:
            return False
        return True


class RunningSupervisor(RunningHelper):

    action = None
    reload_timeout = 60

    namevar = "service"

    def verify(self):
        self.parent.assert_no_changes()
        self.assert_file_is_current(
            self.parent.pidfile, ["bin/supervisord", "etc/supervisord.conf"]
        )
        if not self.is_running():
            raise UpdateNeeded()

    def update(self):
        if not self.is_running():
            self.service.start()
        else:
            self.reload_supervisor()
        if self.parent.wait_for_running:
            # Wait max startup time now that supervisor is back
            time.sleep(self.parent.max_startup_delay)

    def reload_supervisor(self):
        self.cmd("bin/supervisorctl reload")
        # Reload is asynchronous and doesn't wait for supervisor to become
        # fully running again. We actually could monitor supervisorctl,
        # though. This can take a long time if supervisor needs to orderly
        # shut down a lot of services.
        wait = self.reload_timeout
        while wait:
            out, err = "", ""
            # Supervisor tends to "randomly" set zero and non-zero exit codes.
            # See https://github.com/Supervisor/supervisor/issues/24 :-/
            out, err = self.cmd("bin/supervisorctl pid", ignore_returncode=True)
            if "SHUTDOWN_STATE" in out:
                time.sleep(1)
                continue
            try:
                int(out) > 0
            except ValueError:
                time.sleep(1)
                wait -= 1
            else:
                break
        else:
            raise RuntimeError(
                "supervisor master process did not start within {} seconds".format(
                    self.reload_timeout
                )
            )


class StoppedSupervisor(RunningHelper):
    def verify(self):
        if self.is_running():
            raise UpdateNeeded()

    def update(self):
        self.cmd("bin/supervisorctl shutdown")



import json
import os
import sys
import textwrap
from unittest import mock

import pytest

from .. import (
    CONFIG_FILE_NAME,
    assert_up_to_date,
    main,
    migrate,
    read_config,
    write_config,
)


def test_migrate__read_config__1(tmp_path):
    """It returns the migration configuration version from ``.batou.json.``."""
    (tmp_path / CONFIG_FILE_NAME).write_text(
        json.dumps({"migration": {"version": 2300}})
    )
    os.chdir(tmp_path)
    assert read_config() == 2300


def test_migrate__read_config__2(tmp_path):
    """It raises a FileNotFoundError if the configuration file is missing."""
    with pytest.raises(FileNotFoundError):
        read_config()


def test_migrate__read_config__3(tmp_path):
    """It raises a KeyError if the configuration file does not have the ...

    right structure.
    """
    (tmp_path / CONFIG_FILE_NAME).write_text(json.dumps({"version": 2300}))
    os.chdir(tmp_path)
    with pytest.raises(KeyError):
        assert read_config()


def test_migrate__read_config__4(tmp_path):
    """It converts the migration version number to an integer."""
    (tmp_path / CONFIG_FILE_NAME).write_text(
        json.dumps({"migration": {"version": "2411"}})
    )
    os.chdir(tmp_path)
    assert read_config() == 2411


@pytest.fixture(scope="function")
def migrations(tmp_path):
    """Create some simple migrations."""
    TEMPLATE = textwrap.dedent(
        """
        def migrate(output):
            pass
    """
    )
    os.chdir(tmp_path)
    package = tmp_path / "package"
    package.mkdir()
    (package / "__init__.py").touch()
    migrations = package / "migrations"
    migrations.mkdir()
    (migrations / "__init__.py").touch()
    for step in (2411, 2301, 2300, 2299):
        (migrations / f"{step}.py").write_text(TEMPLATE.format(step))
    sys.path.insert(0, str(tmp_path))
    with mock.patch("batou.migrate.MIGRATION_MODULE", new="package"):
        yield
    sys.path.pop()


def test_migrate__migrate__1(migrations, output):
    """It runs all migration steps starting from next after the ...

    given version.
    """
    assert migrate(2300) == 2411
    assert "Version: 2301\n\nVersion: 2411\n\n" == output.backend.output


def test_migrate__migrate__2(migrations, output):
    """It does nothing if we are already on the latest migration step."""
    assert migrate(2411) == 2411
    assert "" == output.backend.output


def test_migrate__write_config__1(tmp_path):
    """It writes the configuration file."""
    os.chdir(tmp_path)
    write_config(2300)
    assert (tmp_path / CONFIG_FILE_NAME).exists()
    assert read_config() == 2300


def test_migrate__write_config__2(tmp_path):
    """It overwrites an existing the configuration file."""
    os.chdir(tmp_path)
    write_config(2300)
    write_config(2311)
    assert read_config() == 2311


def test_migrate__assert_up_to_date__1(migrations):
    """It returns `True` if the expected version matches the current one."""
    write_config(2411)
    assert assert_up_to_date()


def test_migrate__assert_up_to_date__2(migrations, output):
    """It stops the batou run if the expected version does not match ...

    the current one.
    """
    with pytest.raises(SystemExit):
        assert_up_to_date()
    assert (
        "ERROR: Please run `./batou migrate` first.\n" == output.backend.output
    )


def test_migrate__main__1(tmp_path, migrations, capsys):
    """It runs all migrations if no configuration file is present.

    It creates a configuration file.
    """
    main()
    # We explicitly test the output to the TerminalBackend.
    assert (
        capsys.readouterr().out
        == """\
Current version: 0
Version: 2299

Version: 2300

Version: 2301

Version: 2411

Reached version: 2411
"""
    )
    assert (tmp_path / CONFIG_FILE_NAME).exists()
    assert read_config() == 2411


def test_migrate__main__2(migrations, capsys):
    """It runs no migrations if already at the latest migration step."""
    write_config(2411)
    main()
    # We explicitly test the output to the TerminalBackend.
    assert (
        """\
Current version: 2411
Reached version: 2411
"""
        == capsys.readouterr().out
    )
    assert read_config() == 2411


def test_migrate__main__3(migrations, capsys):
    """It runs no migrations if `bootstrap` is `True` but ...

    sets migration version to current.
    """
    main(bootstrap=True)
    # We explicitly test the output to the TerminalBackend.
    assert "" == capsys.readouterr().out
    assert read_config() == 2411



import glob
import os
import shutil
import sys
import textwrap

import pytest

from batou.environment import UnknownEnvironmentError

from ..encryption import GPGEncryptedFile
from ..manage import (
    add_user,
    decrypt_to_stdout,
    reencrypt,
    remove_user,
    summary,
)
from .test_secrets import cleartext_file, encrypted_file


@pytest.mark.parametrize("func", (add_user, remove_user))
def test_manage__1(monkeypatch, func):
    """It raises an exception if called with an unknown environment."""
    monkeypatch.chdir("examples/errors")
    with pytest.raises(UnknownEnvironmentError) as err:
        func("max@example.com", "foo,bar,errors")
    assert "Unknown environment(s): foo, bar" == str(err.value)


def test_manage__2(tmp_path, monkeypatch, capsys):
    """It allows to add/remove users."""
    shutil.copytree("examples/errors", tmp_path / "errors")
    monkeypatch.chdir(tmp_path / "errors")

    summary()
    out, err = capsys.readouterr()
    assert "306151601E813A47" in out

    remove_user("306151601E813A47", "errors")
    summary()
    out, err = capsys.readouterr()
    assert "306151601E813A47" not in out

    add_user("306151601E813A47", "errors")
    summary()
    out, err = capsys.readouterr()
    assert "306151601E813A47" in out


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="age is available in tests with python 3.7 only",
)
def test_manage__2_age(tmp_path, monkeypatch, capsys):
    """It allows to add/remove_users in an age encrypted environment."""
    shutil.copytree("examples/tutorial-secrets", tmp_path / "tutorial-secrets")
    monkeypatch.chdir(tmp_path / "tutorial-secrets")

    key_name = "https://github.com/ctheune.keys"

    summary()
    out, err = capsys.readouterr()
    assert key_name in out

    remove_user(key_name, "age,age-diffable")
    summary()
    out, err = capsys.readouterr()
    assert key_name not in out

    add_user(key_name, "age,age-diffable")
    summary()
    out, err = capsys.readouterr()
    assert key_name in out


def test_manage__summary__1(capsys, monkeypatch):
    """It prints a summary of the environments, members and secret files."""
    monkeypatch.chdir("examples/errors")
    assert summary() == 0  # exit code
    out, err = capsys.readouterr()
    expected = textwrap.dedent(
        """\
        errors
        \t members
        \t\t- 03C7E67FC9FD9364
        \t\t- 306151601E813A47
        \t secret files
        \t\t(none)

    """
    )
    assert out == expected
    assert err == ""


def test_manage__summary__2(capsys, monkeypatch):
    """It prints an error message on decoding problems."""
    monkeypatch.chdir("examples/errors")
    monkeypatch.setitem(os.environ, "GNUPGHOME", "")
    assert summary() == 1  # exit code
    out, err = capsys.readouterr()
    expected = "Exitcode 2 while calling: gpg --decrypt"
    assert out == "errors\n"
    assert err.startswith(expected)


def test_manage__summary__3(capsys, monkeypatch):
    """It can decode files where the host is not present."""
    monkeypatch.chdir("examples/errors2")
    assert summary() == 0  # exit code
    out, err = capsys.readouterr()
    expected = "secretserror\n\t members"
    assert expected in out
    assert err == ""


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="age is available in tests with python 3.7 only",
)
def test_manage__reencrypt__1(tmp_path, monkeypatch, capsys):
    """It re-encrypts all files with the current members."""
    shutil.copytree("examples/tutorial-secrets", tmp_path / "tutorial-secrets")

    monkeypatch.chdir(tmp_path / "tutorial-secrets")

    # read files environments/*/secret*
    # and make sure all of them change
    # when we re-encrypt

    old = {}
    for path in glob.glob("environments/*/secret*"):
        with open(path, "rb") as f:
            old[path] = f.read()

    reencrypt("")  # empty string means all environments
    new = {}
    for path in glob.glob("environments/*/secret*"):
        with open(path, "rb") as f:
            new[path] = f.read()

    for path in old:
        assert old[path] != new[path]

    assert set(old) == set(new)


def test_manage__decrypt_to_stdout__1(encrypted_file, capsys):
    """It decrypts a file and writes the content to stdout."""
    with GPGEncryptedFile(encrypted_file) as secret:
        with open(cleartext_file) as cleartext:
            # assert cleartext.read().strip() == secret.cleartext.strip()
            assert decrypt_to_stdout(str(encrypted_file)) == 0
            out, err = capsys.readouterr()
            assert out == cleartext.read()
            assert err == ""



import os.path
import pathlib
import sys

import mock
import pytest

from batou.utils import CmdExecutionError

from ..buildout import Buildout
from ..file import File


def buildout(**kw):
    buildout = Buildout(**kw)
    buildout.cmd = mock.Mock()
    return buildout


def test_update_should_pass_config_file_name(root):
    b = buildout(
        python="2.7", setuptools="1.0", config=File("myown.cfg", content="")
    )
    root.component += b
    b.update()

    assert b.cmd.call_count == 1
    calls = iter(x[1][0] for x in b.cmd.mock_calls)
    assert next(calls) == 'bin/buildout -t 3 -c "myown.cfg"'


def test_update_should_pass_custom_timeout(root):
    b = buildout(python="2.7", setuptools="1.0", timeout=40)
    b.update()

    assert b.cmd.call_count == 1
    calls = iter(x[1][0] for x in b.cmd.mock_calls)
    assert next(calls) == 'bin/buildout -t 40 -c "buildout.cfg"'


@pytest.mark.slow
@pytest.mark.timeout(60)
@pytest.mark.skipif(
    sys.version_info >= (3, 7),
    reason="python 2.7 only available in tests with python 3.6",
)
def test_runs_buildout_successfully(root):
    b = Buildout(
        python="2.7",
        version="2.13.4",
        setuptools="44.1.1",
        config=File(
            "buildout.cfg",
            source=pathlib.Path(__file__).parent / "buildout-example.cfg",
        ),
    )
    root.component += b
    root.component.deploy()
    assert os.path.isdir(
        os.path.join(root.environment.workdir_base, "mycomponent/parts")
    )
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/bin/py")
    )


@pytest.mark.slow
@pytest.mark.timeout(60)
def test_runs_buildout3_successfully(root, output):
    b = Buildout(
        python="3",
        version="3.0.1",
        setuptools="68.0.0",
        wheel="0.36.2",
        pip="23.3.2",
        config=File(
            "buildout3.cfg",
            source=str(pathlib.Path(__file__).parent / "buildout3-example.cfg"),
        ),
    )
    root.component += b
    try:
        root.component.deploy()
    except CmdExecutionError as e:
        e.report()
        print(output.backend.output)
        raise
    assert os.path.isdir(
        os.path.join(root.environment.workdir_base, "mycomponent/parts")
    )
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/bin/py")
    )



import os.path

from batou.component import Component
from batou.lib.archive import Extract
from batou.lib.download import Download


class Configure(Component):
    # XXX not convergent when changing args

    namevar = "path"
    args = ""
    prefix = None
    build_environment = None

    def configure(self):
        if self.prefix is None:
            self.prefix = self.workdir

    def verify(self):
        # This is guesswork. Unfortunately CMMI doesn't work any better.
        self.assert_file_is_current(self.path + "/config.status")
        self.assert_file_is_current(
            self.path + "/.batou.config.success", [self.path + "/configure"]
        )

    def update(self):
        with self.chdir(self.path):
            self.cmd(
                self.expand(
                    "./configure --prefix={{component.prefix}} "
                    "{{component.args}}"
                ),
                env=self.build_environment,
            )
            self.touch(".batou.config.success")


class Make(Component):

    namevar = "path"
    build_environment = None

    def verify(self):
        self.assert_file_is_current(
            self.path + "/.batou.make.success", [self.path + "/Makefile"]
        )

    def update(self):
        with self.chdir(self.path):
            self.cmd("make install", env=self.build_environment)
            self.touch(".batou.make.success")


class Build(Component):
    """Complex build definition for

    - downloading
    - extracting
    - configure
    - make install

    """

    namevar = "uri"
    checksum = None
    configure_args = ""
    prefix = None
    build_environment = None

    def configure(self):
        download = Download(self.uri, checksum=self.checksum)
        self += download

        extract = Extract(download.target, strip=1)
        self += extract

        self += Configure(
            extract.target,
            args=self.configure_args,
            prefix=self.prefix,
            build_environment=self.build_environment,
        )
        self += Make(extract.target, build_environment=self.build_environment)

    @property
    def namevar_for_breadcrumb(self):
        return os.path.basename(self.uri)



# -*- coding: utf-8 -*-
import os
import pathlib
import sys
import time

import pytest

from batou.lib.archive import Extract


def test_unknown_extension_raises():
    extract = Extract("example.unknown")
    extract.workdir = ""
    with pytest.raises(ValueError):
        extract.configure()


def test_missing_archive_predicts_change(root):
    extract = Extract("example.tar.gz")
    root.component += extract

    extractor = extract.sub_components[0]
    with pytest.raises(AssertionError):
        extractor.verify()


def test_untar_extracts_archive_to_target_directory(root):
    extract = Extract(
        # resource_filename(__name__, "example.tar.gz"), target="example"
        pathlib.Path(__file__).parent / "example.tar.gz",
        target="example",
    )
    root.component += extract
    root.component.deploy()
    assert os.listdir(str(extract.target)) == ["foo"]


def test_ignores_ctime_for_directories(root):
    # This is hard to test: ctime can not be changed directly,
    # we thus have to perform a somewhat elaborate dance to align
    # the starts as we wish.
    # archive = resource_filename(__name__, "example.tar.gz")
    archive = pathlib.Path(__file__).parent / "example.tar.gz"
    extract = Extract(archive, target="example")

    root.component += extract
    root.component.deploy()
    # No assertion raised, nothing to extract
    extract.extractor.verify()

    # Make the archive's ctime newer than the directories, but ensure
    # that the file is current (so only the archives _would_ trigger)
    # if we didn't filter them out properly.
    time.sleep(1.1)
    now = (time.time(), time.time())
    os.utime(archive, now)
    os.utime(str(extract.target) + "/foo/bar/qux", now)

    # No assertion raised, still nothing to extract
    extract.extractor.verify()


def test_untar_can_strip_paths_off_archived_files(root):
    extract = Extract(
        # resource_filename(__name__, "example.tar.gz"), target="example", strip=1
        pathlib.Path(__file__).parent / "example.tar.gz",
        target="example",
        strip=1,
    )
    root.component += extract
    root.component.deploy()
    assert os.listdir(str(extract.target)) == ["bar"]


def test_zip_extracts_archive_to_target_directory(root):
    extract = Extract(
        # resource_filename(__name__, "example.zip"), target="example"
        pathlib.Path(__file__).parent / "example.zip",
        target="example",
    )
    root.component += extract
    root.component.deploy()
    assert os.listdir(str(extract.target)) == ["foo"]


def test_zip_overwrites_existing_files(root):
    extract = Extract(
        # resource_filename(__name__, "example.zip"), target="example"
        pathlib.Path(__file__).parent / "example.zip",
        target="example",
    )
    root.component += extract

    target = "%s/mycomponent/example/foo/bar" % root.environment.workdir_base
    os.makedirs(target)
    filename = target + "/qux"
    with open(filename, "w") as f:
        f.write("foo")
    # Bypass verify (which would do nothing since the just written files
    # are newer than the zip file). XXX The API is a little kludgy here.
    extract.sub_components[0].update()
    with open(filename) as f:
        assert "" == f.read()


@pytest.mark.slow
@pytest.mark.skipif(sys.platform != "darwin", reason="only runs on OS X")
def test_dmg_extracts_archive_to_target_directory(root):
    extract = Extract(
        # resource_filename(__name__, "example.dmg"), target="example"
        pathlib.Path(__file__).parent / "example.dmg",
        target="example",
    )
    root.component += extract
    root.component.deploy()

    assert sorted(os.listdir(str(extract.target))) == [
        " ",
        "a\u0308sdf.txt",
        "example.app",
    ]

    # ' ' is a symlink which stays one after copying:
    assert os.path.islink(extract.target + "/ ")
    start_bin = extract.target + "/example.app/MacOS/start.bin"
    with open(start_bin) as start_bin:
        assert start_bin.read() == "I start the example app! ;)"


def test_dmg_does_not_support_strip(root):
    extract = Extract(
        # resource_filename(__name__, "example.dmg"), strip=1, target="example"
        pathlib.Path(__file__).parent / "example.dmg",
        strip=1,
        target="example",
    )
    with pytest.raises(ValueError) as e:
        root.component += extract
        assert e.value.args[0] == "Strip is not supported by DMGExtractor"



import os
import os.path

import pytest
from mock import Mock

import batou
from batou import SilentConfigurationError, UpdateNeeded
from batou.component import (
    Attribute,
    Component,
    ConfigString,
    RootComponent,
    handle_event,
    platform,
)
from batou.utils import CmdExecutionError


@pytest.fixture
def mockroot(tmpdir):
    return Mock(overrides={}, defdir=str(tmpdir))


class SampleComponent(Component):
    updated = False
    needs_update = True

    def verify(self):
        assert not self.needs_update

    def update(self):
        self.updated = True


def test_init_with_no_arguments_creates_plain_component():
    component = Component()
    # Lazy initialized attribute
    assert not hasattr(component, "sub_components")


def test_plain_component_runs_noop_configure_verify_update():
    component = Component()
    component.configure()
    component.verify()
    component.update()


def test_component_with_unknown_arguments_raises_valueerror():
    # Regression against bug #65
    with pytest.raises(ValueError) as e:
        Component("foo")
    assert str(e.value) == "Namevar is undefined for this component."
    with pytest.raises(ValueError) as e:
        Component(asdf="foo")
    assert str(e.value) == (
        "The following arguments are unacceptable for this component "
        "because they are either undefined or would override "
        "methods: asdf"
    )


def test_init_component_with_namevar_uses_first_argument():
    class SampleComponent(Component):
        namevar = "asdf"

    component = SampleComponent("foobar")
    assert "foobar" == component.asdf


def test_init_component_with_namevar_fails_without_argument():
    class SampleComponent(Component):
        namevar = "asdf"

    with pytest.raises(ValueError):
        SampleComponent()


def test_init_keyword_args_update_dict():
    class TestComponent(Component):
        foobar = None

    component = TestComponent(foobar=1)
    assert 1 == component.foobar


def test_prepare_sets_up_vars(root):
    assert root.component.environment is root.environment
    assert root.component.host is root.host
    assert root.component.root is root
    assert root.component.parent is root


def test_op_orassignment_ignores_none(root):
    root.component |= None


def test_recursive_sub_component_iterator(root):
    class TestComponent(Component):
        name = None

    for x in range(3):
        c = TestComponent(name="x{}".format(x))
        root.component += c
        for y in range(2):
            c2 = TestComponent(name="x{}y{}".format(x, y))
            c += c2

    recursive = list(x.name for x in root.component.recursive_sub_components)
    assert [
        "x0",
        "x0y0",
        "x0y1",
        "x1",
        "x1y0",
        "x1y1",
        "x2",
        "x2y0",
        "x2y1",
    ] == recursive


def test_op_orassignment_ignores_already_preapred_component(root):
    class Counting(Component):
        x = 1

        def configure(self):
            self.x += 1

    c = Counting()
    root.component |= c
    assert c.x == 2
    root.component |= c
    assert c.x == 2


def test_prepare_calls_configure(mockroot):
    class TestComponent(Component):

        cwd = None

        def configure(self):
            self.cwd = os.getcwd()

    component = TestComponent()
    assert component.cwd is None
    os.chdir("/tmp")
    component.prepare(mockroot)

    assert component.cwd == mockroot.defdir


def test_prepare_configures_applicable_platforms_as_subcomponents(root):
    class MyComponent(Component):
        pass

    class MyOtherComponent(MyComponent):
        pass

    @platform("testplatform", MyComponent)
    class MyPlatform(Component):
        pass

    @platform("otherplatform", MyComponent)
    class MyOtherPlatform(Component):
        pass

    root.host.platform = "testplatform"
    component = MyComponent()
    root.component += component
    assert 1 == len(component.sub_components)
    assert isinstance(component.sub_components[0], MyPlatform)
    # Because we currently have no defined behaviour about this
    # I'm playing this rather safe: a sub-class of a component does not
    # automatically get to have the same platform components applied.
    other_component = MyOtherComponent()
    root.component += other_component
    assert 0 == len(other_component.sub_components)


def test_deploy_empty_component_runs_without_error(root):
    root.component.deploy()


def test_deploy_update_performed_if_needed(root):
    component = SampleComponent(needs_update=True)
    root.component += component
    root.component.deploy()
    assert component.updated


def test_deploy_update_not_performed_if_not_needed(root):
    component = SampleComponent(needs_update=False)
    root.component += component
    root.component.deploy()
    assert not component.updated


def test_sub_components_are_deployed_first(root):
    log = []

    class MyComponent(Component):
        namevar = "id"

        def verify(self):
            log.append("{}:verify".format(self.id))
            raise batou.UpdateNeeded()

        def update(self):
            log.append("{}:update".format(self.id))

    c1 = MyComponent("1")
    root.component += c1
    c1 += MyComponent("2")
    root.component.deploy()
    assert ["2:verify", "2:update", "1:verify", "1:update"] == log


def test_log_in_configure_is_stored_on_root_for_later(root):
    class MyComponent(Component):

        namevar = "id"

        def configure(self):
            self.log("{}: configure".format(self.id))

    c1 = MyComponent("1")
    root.component += c1
    assert [("1: configure", ())] == c1.root._logs


def test_log_in_verify(root):
    class MyComponent(Component):

        namevar = "id"

        def verify(self):
            self.log("{}: verify".format(self.id))

    c1 = MyComponent("1")
    root.component += c1
    root.log_finish_configure()
    c1.verify()
    assert c1.root._logs is None


def test_adding_subcomponents_configures_them_immediately(root):
    class MyComponent(Component):
        configured = False

        def configure(self):
            self.configured = True

    my = MyComponent()
    root.component += my
    assert my.configured


def test_adding_subcomponents_makes_them_available_as_underscore(root):
    c = Component()
    assert root.component._ is None
    root.component += c
    assert root.component._ is c


def test_oring_subcomponents_makes_them_available_as_underscore(root):
    c = Component()
    assert root.component._ is None
    root.component |= c
    assert root.component._ is c


# AFIC = assert file is current
def test_afic_raises_if_nonexisting_file(root):
    component = Component()
    root.component += component
    with pytest.raises(batou.UpdateNeeded):
        component.assert_file_is_current("idonotexist")


def test_afic_doesnt_raise_if_file_exists_but_no_reference_is_given(root):
    component = Component()
    root.component += component
    component.assert_file_is_current(__file__)


def test_afic_raises_if_file_isolder_than_reference(tmpdir, root):
    component = Component()
    root.component += component
    with pytest.raises(batou.UpdateNeeded):
        component.assert_file_is_current(__file__, [str(tmpdir)])


# ANSC = assert no subcomponent changes


def test_ansc_raises_if_subcomponent_changed(root):
    c2 = Component()
    root.component += c2
    c2.changed = True
    with pytest.raises(batou.UpdateNeeded):
        root.component.assert_no_subcomponent_changes()


def test_ansc_does_not_raise_if_no_subcomponent_changed(root):
    c2 = Component()
    root.component += c2
    c2.changed = False
    root.component.assert_no_subcomponent_changes()


# ACIC = assert component is current


def test_acic_raises_if_no_reference(root):
    c = Component()
    c.last_updated = Mock(return_value=None)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    with pytest.raises(batou.UpdateNeeded):
        c.assert_component_is_current(c2)


def test_acic_raises_if_older_reference(root):
    c = Component()
    c.last_updated = Mock(return_value=20)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    with pytest.raises(batou.UpdateNeeded):
        c.assert_component_is_current(c2)


def test_acic_does_not_raise_if_current(root):
    c = Component()
    c.last_updated = Mock(return_value=21)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    c.assert_component_is_current(c2)


def test_acic_does_not_raise_if_newer(root):
    c = Component()
    c.last_updated = Mock(return_value=22)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    c.assert_component_is_current(c2)


def test_acic_accepts_multiple_components(root):
    c = Component()
    c.last_updated = Mock(return_value=20)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    with pytest.raises(batou.UpdateNeeded):
        c.assert_component_is_current([c2, c2])


def test_cmd_expands_jinja(mockroot):
    c = Component()
    c.foo = "asdf"
    c.prepare(mockroot)
    assert ("asdf\n", "") == c.cmd('echo "{{component.foo}}"')


def test_cmd_returns_output(mockroot):
    c = Component()
    c.prepare(mockroot)
    assert ("1\n", "") == c.cmd("echo 1")


def test_cmd_raises_if_error(mockroot):
    c = Component()
    c.prepare(mockroot)
    with pytest.raises(CmdExecutionError):
        c.cmd("non-existing-command")


def test_cmd_returns_output_if_ignore_returncode(mockroot):
    c = Component()
    c.prepare(mockroot)
    out, err = c.cmd("echo important output && false", ignore_returncode=True)
    assert "important output\n" == out


def test_touch_creates_new_file(tmpdir):
    reference = str(tmpdir / "reference")
    assert not os.path.exists(reference)
    c = Component()
    c.touch(reference)
    assert os.path.exists(reference)


def test_touch_updates_mtime_leaves_content_intact(tmpdir):
    reference = str(tmpdir / "reference")
    with open(reference, "w") as r:
        r.write("Hello world")
    c = Component()
    os.utime(reference, (0, 0))
    c.touch(reference)
    assert os.stat(reference).st_mtime != 0
    with open(reference, "r") as r:
        assert "Hello world" == r.read()


def test_expand(root):
    assert root.component.expand("Hello {{host.fqdn}}") == ("Hello localhost")


def test_templates(root):
    with open("sample", "w") as template:
        template.write("Hello {{host.fqdn}}")
    assert root.component.template("sample") == ("Hello localhost\n")


def test_template_batou_generated_header(root):
    with open("sample", "w") as template:
        template.write("Hello\n# {{batou_generated_header}}")
    assert root.component.template("sample") == (
        "Hello\n# File is generated by batou for component MyComponent. Don't edit manually.\n"
    )


def test_chdir_contextmanager_is_stackable():
    outer = os.getcwd()
    inner1 = os.path.join(os.path.dirname(__file__), "fixture")
    inner2 = os.path.join(os.path.dirname(__file__))
    c = Component()
    with c.chdir(inner1):
        assert inner1 == os.getcwd()
        with c.chdir(inner2):
            assert inner2 == os.getcwd()  # noqa
        assert inner1 == os.getcwd()
    assert outer == os.getcwd()


def test_root_and_component_know_working_dir(root):
    assert root.workdir.endswith("/work/mycomponent")
    assert root.component.workdir.endswith("/work/mycomponent")


def test_root_component_repr():
    host = Mock()
    environment = Mock()
    root = RootComponent(
        "haproxy", environment, host, [], False, object, "/defdir", "/workdir"
    )
    assert repr(root).startswith('<RootComponent "haproxy" object at ')


def test_component_manages_working_dir(root):
    previous_workdir = os.getcwd()
    os.rmdir(root.workdir)

    class RememberWorkDir(Component):
        def verify(self):
            self.parent.i_was_in = os.getcwd()

    root.component += RememberWorkDir()
    root.component.deploy()
    assert os.path.isdir(root.workdir)
    assert os.path.realpath(root.component.i_was_in) == os.path.realpath(
        root.workdir
    )
    cwd = os.getcwd()
    # Need to ensure we resolve symlinks: on OS x /var/tmp may lead to
    # /private/...
    assert os.path.realpath(cwd) == os.path.realpath(previous_workdir)


def test_cmd_execution_failed_gives_command_in_exception(mockroot):
    c = Component()
    c.prepare(mockroot)
    with pytest.raises(CmdExecutionError) as e:
        c.cmd("asdf")
    assert e.value.cmd == "asdf"
    assert e.value.returncode == 127


def test_cmd_should_not_stop_if_process_expects_input(mockroot):
    c = Component()
    c.prepare(mockroot)
    stdout, stderr = c.cmd("cat")
    # The assertion is, that the test doesn't get stuck .


def test_last_updated_not_implemented_on_base():
    c = Component()
    with pytest.raises(NotImplementedError):
        c.last_updated()


def test_require_one_convenience_api_returns_scalar():
    c = Component()
    c.require = Mock(return_value=[1])
    assert 1 == c.require_one("asdf")


def test_require_one_convenience_raises_if_no_result():
    c = Component()
    c.require = Mock(return_value=[])
    with pytest.raises(SilentConfigurationError):
        c.require_one("asdf")


def test_require_one_convenience_raises_if_more_results():
    c = Component()
    c.require = Mock(return_value=[1, 2])
    with pytest.raises(KeyError):
        c.require_one("asdf")


def test_assert_cmd_when_succesful(mockroot):
    c = Component()
    c.prepare(mockroot)
    c.assert_cmd("true")


def test_assert_cmd_when_unsuccessful(mockroot):
    c = Component()
    c.prepare(mockroot)
    with pytest.raises(UpdateNeeded):
        c.assert_cmd("false")


def test_assert_no_changes_local_does_not_raise(root):
    root.component.assert_no_changes()


def test_assert_no_changes_local_raises(root):
    root.component.changed = True
    with pytest.raises(batou.UpdateNeeded):
        root.component.assert_no_changes()


def test_assert_no_changes_recursive_does_not_raise(root):
    root.component += Component()
    root.component.assert_no_changes()


def test_assert_no_changes_recursive_raises(root):
    c2 = Component()
    root.component += c2
    c2.changed = True
    with pytest.raises(batou.UpdateNeeded):
        root.component.assert_no_changes()


def test_root_overrides_missing_attribute_raises(root):
    root.overrides = {"asdf": 1}
    with pytest.raises(batou.MissingOverrideAttributes) as e:
        root.prepare()
    assert e.value.attributes == ["asdf"]


def test_root_overrides_existing_attribute(root):
    root.factory.asdf = None
    root.overrides = {"asdf": 1}
    root.prepare()
    assert root.component.asdf == 1


@pytest.mark.parametrize(
    "conv_func,conf_value,expected",
    [
        ("list", ConfigString(""), []),
        ("list", ConfigString("1,2"), ["1", "2"]),
        ("list", ConfigString("3"), ["3"]),
        ("list", ConfigString("  3, 3,"), ["3", "3"]),
        ("list", [], []),
        ("literal", ConfigString("[3,3]"), [3, 3]),
        ("literal", ConfigString("True"), True),
        ("literal", ConfigString("False"), False),
    ],
)
def test_attribute_conversion_functions(conv_func, conf_value, expected, root):
    """It converts config strings with conversion functions."""

    class Foo(Component):
        a = Attribute(conv_func, default=conf_value)

    f = Foo()
    root.component += f
    assert f.a == expected


def test_attribute_missing_override_and_default(root):
    class Foo(Component):
        a = Attribute(str)

        def configure(self):
            self.a

    f = Foo()
    with pytest.raises(AttributeError) as e:
        root.component += f

    assert (
        e.value.args[0]
        == 'No override and no default given for `a` on Foo instance <Foo (localhost) "MyComponent > Foo">.'
    )


@pytest.mark.parametrize(
    "default,expected",
    [
        (["3"], ["3"]),
        ("True", "True"),
        (True, True),
    ],
)
def test_attribute_conversion_default(default, expected, root):
    """It does not convert default values with conversion function."""

    class Foo(Component):
        a = Attribute("list", default=default)

    f = Foo()
    root.component += f
    assert f.a == expected


class EventHandlingComponent(Component):

    namevar = "x"
    perform_update = True

    def verify(self):
        if self.perform_update:
            raise UpdateNeeded()

    def configure(self):
        self.log = []

    @handle_event("before-update", "*")
    def handle(self, component):
        self.log.append(component.x)


def test_event_handler_before_update_no_changes(root):
    c = EventHandlingComponent("1", perform_update=False)
    root.component += c
    root.component.deploy()
    assert c.log == []


def test_event_handler_before_update_with_changes_all(root):
    c1 = EventHandlingComponent("1")
    c2 = EventHandlingComponent("2")
    root.component += c1
    root.component += c2
    root.component.deploy()
    assert c1.log == ["1", "2"]
    assert c2.log == ["1", "2"]


def test_event_handler_before_update_with_changes_precursor(root):
    from batou.component import handle_event

    log = []

    class Foo(Component):
        namevar = "x"

        def verify(self):
            raise UpdateNeeded()

        @handle_event("before-update", "precursor")
        def handle(self, component):
            log.append((self.x, component.x))

    root.component += Foo("1")
    root.component += Foo("2")
    root.component.deploy()
    assert log == [("2", "1")]


def test_checksum_returns_even_when_never_a_value_was_passed():
    c = SampleComponent()
    assert (
        "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
        == c.checksum()
    )


def test_checksum_updates_and_returns():
    c = SampleComponent()
    assert (
        "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae"
        == c.checksum(b"foo")
    )


def test_checksum_depends_on_order():
    c = SampleComponent()
    c.checksum(b"foo")
    c.checksum(b"bar")
    assert (
        "c3ab8ff13720e8ad9047dd39466b3c8974e592c2fa383d4a3960714caef0c4f2"
        == c.checksum()
    )

    c = SampleComponent()
    c.checksum(b"bar")
    c.checksum(b"foo")
    assert (
        "88ecde925da3c6f8ec3d140683da9d2a422f26c1ae1d9212da1e5a53416dcc88"
        == c.checksum()
    )


def test_checksum_multiple_calls_do_not_change_checksum():
    c = SampleComponent()
    c.checksum(b"foo")
    c.checksum(b"bar")
    assert (
        "c3ab8ff13720e8ad9047dd39466b3c8974e592c2fa383d4a3960714caef0c4f2"
        == c.checksum()
    )
    assert (
        "c3ab8ff13720e8ad9047dd39466b3c8974e592c2fa383d4a3960714caef0c4f2"
        == c.checksum()
    )



import sys

import pytest

from batou.component import Component
from batou.lib.python import VirtualEnv


@pytest.mark.skipif(
    sys.version_info >= (3, 7),
    reason="python 2.7 only available in tests with python 3.6",
)
def test_venv_updates_if_python_changes(root):
    import ast

    class Playground(Component):
        namevar = "version"

        def configure(self):
            self.venv = VirtualEnv(self.version)
            self += self.venv

    playground = Playground("2.7")
    root.component += playground
    playground.deploy()
    root.component.sub_components.remove(playground)

    playground = Playground("3")
    root.component += playground
    playground.deploy()

    out, err = playground.cmd(
        '{}/bin/python -c "import sys; print(sys.version_info[:2])"'.format(
            playground.workdir
        )
    )
    assert 3 == ast.literal_eval(out)[0]


@pytest.mark.skipif(
    sys.version_info >= (3, 7),
    reason="python 2.7 only available in tests with python 3.6",
)
def test_venv_does_not_update_if_python_does_not_change(root):
    class Playground(Component):
        namevar = "version"

        def configure(self):
            self.venv = VirtualEnv(self.version)
            self += self.venv

    playground = Playground("2.7")
    root.component += playground
    playground.deploy()
    assert playground.changed
    playground.deploy()
    assert not playground.changed



import json
import os
import subprocess
import sys
import tempfile

import execnet

from batou import DeploymentError, RepositoryDifferentError, output
from batou.utils import CmdExecutionError
from batou.utils import cmd as cmd_


def cmd(c, *args, **kw):
    return cmd_("LANG=C LC_ALL=C LANGUAGE=C {}".format(c), *args, **kw)


def find_line_with(prefix, output):
    for line in output.splitlines():
        line = line.strip()
        if line.startswith(prefix):
            return line.replace(prefix, "", 1).strip()


class Repository(object):
    """A repository containing the batou deployment.

    The actual deployment may be located within a prefix
    of this repository. Where the repository starts can be
    determined by the specific repository implementation.

    """

    def __init__(self, environment):
        self.environment = environment
        # We can't set this default on the environment because we
        # have a special use of None for test support.
        self.root = environment.repository_root or "."

    @classmethod
    def from_environment(cls, environment):
        if environment.connect_method == "local":
            return NullRepository(environment)
        elif environment.update_method == "rsync":
            return RSyncRepository(environment)
        elif environment.update_method == "rsync-ext":
            return RSyncExtRepository(environment)
        elif environment.update_method == "hg-bundle":
            return MercurialBundleRepository(environment)
        elif environment.update_method == "hg-pull":
            return MercurialPullRepository(environment)
        elif environment.update_method == "git-bundle":
            return GitBundleRepository(environment)
        elif environment.update_method == "git-pull":
            return GitPullRepository(environment)
        raise ValueError("Could not find method to transfer the repository.")

    def verify(self):
        pass

    def update(self, host):
        pass


class NullRepository(Repository):
    """A repository that does nothing to verify or update."""


class FilteredRSync(execnet.RSync):
    """Implement a filtered RSync that
    avoids copying files from our blacklist.
    """

    IGNORE_LIST = (
        ".appenv",
        ".batou",
        ".batou-lock",
        ".git",
        ".hg",
        ".kitchen",
        ".vagrant",
        "work",
    )

    def __init__(self, *args, **kw):
        super(FilteredRSync, self).__init__(*args, **kw)
        self.IGNORE_LIST = set(self.IGNORE_LIST)

    def filter(self, path):
        return os.path.basename(path) not in self.IGNORE_LIST


class RSyncRepository(Repository):
    def verify(self):
        output.annotate(
            "You are using rsync. This is a non-verifying repository "
            "-- continuing on your own risk!",
            red=True,
        )

    def update(self, host):
        source, target = self.root, host.remote_repository
        output.annotate("rsync: {} -> {}".format(source, target), debug=True)
        rsync = FilteredRSync(source, verbose=False)
        # We really want to use `delete=True` here but there's an execnet issue
        # preventing us to use it. See
        # https://github.com/flyingcircusio/batou/issues/107
        rsync.add_target(host.gateway, target)
        rsync.send()


class RSyncExtRepository(Repository):
    IGNORE_LIST = (
        ".appenv",
        ".batou",
        ".batou-lock",
        ".git",
        ".hg",
        ".kitchen",
        ".vagrant",
        "work",
    )

    SYNC_OPTS = [
        "--recursive",
        "--links",
        "--delete",
    ]

    def verify(self):
        output.annotate(
            "You are using external rsync. This is a non-verifying "
            "repository -- continuing on your own risk!",
            red=True,
        )

    def update(self, host):
        source, dest = self.root, host.remote_repository

        rsync_args = self.SYNC_OPTS.copy()

        ssh_configs = [
            "ssh_config_{}".format(self.environment.name),
            "ssh_config",
        ]
        for ssh_config in ssh_configs:
            if os.path.exists(ssh_config):
                rsync_args.append(f"--rsh='ssh -F {ssh_config}'")
                break

        for ignore in self.IGNORE_LIST:
            rsync_args.append(f"--exclude='{ignore}'")

        if host.require_sudo:
            rsync_args.append(
                "--rsync-path='sudo -ni -u {} rsync'".format(host.service_user)
            )

        output.annotate("rsync-ext: {} -> {}".format(source, dest), debug=True)

        cmd(
            "rsync {opts} {source}/ {target}:{dest}".format(
                opts=" ".join(rsync_args),
                source=source,
                target=host.fqdn,
                dest=dest,
            )
        )


def hg_cmd(hgcmd):
    output, _ = cmd(hgcmd + " -Tjson")
    output = json.loads(output)
    return output


class MercurialRepository(Repository):

    root = None
    _upstream = None

    def __init__(self, environment):
        super(MercurialRepository, self).__init__(environment)
        self.root = hg_cmd("hg root")[0]["reporoot"]
        self.branch = environment.branch or "default"
        self.subdir = os.path.relpath(self.environment.base_dir, self.root)

    @property
    def upstream(self):
        if self.environment.repository_url is not None:
            self._upstream = self.environment.repository_url
        elif self._upstream is None:
            for item in hg_cmd("hg showconfig paths"):
                if item["name"] != "paths.default":
                    continue
                self._upstream = item["value"]
                break
            else:
                raise AssertionError("`paths.default` not found")
        return self._upstream

    def update(self, host):
        self._ship(host)
        remote_id = host.rpc.hg_update_working_copy(self.branch)
        local_id = hg_cmd("hg id")[0]["id"]
        if self.environment.deployment.dirty:
            local_id = local_id.replace("+", "")
        if remote_id != local_id:
            raise RepositoryDifferentError.from_context(local_id, remote_id)

    def verify(self):
        # Safety belt that we're acting on a clean repository.
        if self.environment.deployment.dirty:
            output.annotate(
                "You are running a dirty deployment. This can cause "
                "inconsistencies -- continuing on your own risk!",
                red=True,
            )
            return

        try:
            status = hg_cmd("hg stat")
        except CmdExecutionError:
            output.error(
                "Unable to check repository status. "
                "Is there an HG repository here?"
            )
            raise
        else:
            if status:
                output.error("Your repository has uncommitted changes.")
                output.annotate(
                    """\
I am refusing to deploy in this situation as the results will be unpredictable.
Please commit and push first.
""",
                    red=True,
                )
                for item in status:
                    output.annotate(
                        "{} {}".format(item["status"], item["path"]), red=True
                    )
                raise DeploymentError("Uncommitted changes")
        try:
            cmd("hg -q outgoing -l 1", acceptable_returncodes=[1])
        except CmdExecutionError:
            output.error(
                """\
Your repository has outgoing changes.

I am refusing to deploy in this situation as the results will be unpredictable.
Please push first.
"""
            )
            raise DeploymentError("Outgoing changes")


class MercurialPullRepository(MercurialRepository):
    def _ship(self, host):
        host.rpc.hg_pull_code(upstream=self.upstream)


class MercurialBundleRepository(MercurialRepository):
    def _ship(self, host):
        heads = host.rpc.hg_current_heads()
        if not heads:
            raise ValueError(
                "Remote repository did not find any heads. "
                "Can not continue creating a bundle."
            )
        fd, bundle_file = tempfile.mkstemp()
        os.close(fd)
        bases = " ".join("--base {}".format(x) for x in heads)
        cmd(
            "hg -qy bundle {} {}".format(bases, bundle_file),
            acceptable_returncodes=[0, 1],
        )
        change_size = os.stat(bundle_file).st_size
        if not change_size:
            return
        output.annotate(
            "Sending {} bytes of changes".format(change_size), debug=True
        )
        rsync = execnet.RSync(bundle_file, verbose=False)
        rsync.add_target(
            host.gateway, host.remote_repository + "/batou-bundle.hg"
        )
        rsync.send()
        os.unlink(bundle_file)
        output.annotate("Unbundling changes", debug=True)
        host.rpc.hg_unbundle_code()


class GitRepository(Repository):

    root = None
    _upstream = None
    remote = "origin"

    def __init__(self, environment):
        super(GitRepository, self).__init__(environment)
        self.branch = environment.branch or "master"
        root = subprocess.check_output(
            ["git", "rev-parse", "--show-toplevel"]
        ).strip()
        self.root = root.decode(sys.getfilesystemencoding())
        self.subdir = os.path.relpath(self.environment.base_dir, self.root)

    @property
    def upstream(self):
        if self.environment.repository_url is not None:
            self._upstream = self.environment.repository_url
        elif self._upstream is None:
            result = cmd("git remote show -n {}".format(self.remote))[0]
            self._upstream = find_line_with("Fetch URL:", result)
        return self._upstream

    def update(self, host):
        self._ship(host)
        remote_id = host.rpc.git_update_working_copy(self.branch)
        # This can theoretically fail if we have a fresh repository, but
        # that doesn't make sense at this point anyway.
        local_id, _ = cmd("git rev-parse HEAD")
        local_id = local_id.strip()
        if remote_id != local_id:
            raise RepositoryDifferentError.from_context(local_id, remote_id)

    def verify(self):
        # Safety belt that we're acting on a clean repository.
        if self.environment.deployment.dirty:
            output.annotate(
                "You are running a dirty deployment. This can cause "
                "inconsistencies -- continuing on your own risk!",
                red=True,
            )
            return

        try:
            status, _ = cmd("git status --porcelain")
        except CmdExecutionError:
            output.error(
                "Unable to check repository status. "
                "Is there a Git repository here?"
            )
            raise
        else:
            status = status.strip()
            if status.strip():
                output.error("Your repository has uncommitted changes.")
                output.annotate(
                    """\
I am refusing to deploy in this situation as the results will be unpredictable.
Please commit and push first.
""",
                    red=True,
                )
                output.annotate(status, red=True)
                raise DeploymentError()
        outgoing, _ = cmd(
            "git log {remote}/{branch}..{branch} --pretty=oneline".format(
                remote=self.remote, branch=self.branch
            ),
            acceptable_returncodes=[0, 128],
        )
        if outgoing.strip():
            output.error(
                """\
Your repository has outgoing changes on branch {branch}:

{outgoing}

I am refusing to deploy in this situation as the results will be unpredictable.
Please push first.
""".format(
                    branch=self.branch, outgoing=outgoing
                )
            )
            raise DeploymentError()


class GitPullRepository(GitRepository):
    def _ship(self, host):
        host.rpc.git_pull_code(upstream=self.upstream, branch=self.branch)


class GitBundleRepository(GitRepository):
    def _ship(self, host):
        head = host.rpc.git_current_head()
        if head:
            # check if head is in local branch, if not, bundle everything and ignore remote head
            try:
                cmd(
                    f"git branch {self.branch} --contains {head}",
                    acceptable_returncodes=[0],
                )
            except CmdExecutionError:
                head = None
        if head is None:
            bundle_range = self.branch
        else:
            head = head.decode("ascii")
            bundle_range = f"{head}..{self.branch}"
        fd, bundle_file = tempfile.mkstemp()
        os.close(fd)
        try:
            out, err = cmd(
                "git bundle create {file} {range}".format(
                    file=bundle_file, range=bundle_range
                ),
                acceptable_returncodes=[0],
            )
        except CmdExecutionError as e:
            if (
                e.returncode == 128
                and "fatal: Refusing to create empty bundle." in e.stderr
            ):
                return
            raise

        change_size = os.stat(bundle_file).st_size
        if change_size == 0:
            output.error("Created invalid bundle (0 bytes):")
            output.annotate(err, red=True)
            raise DeploymentError()
        output.annotate(
            "Sending {} bytes of changes".format(change_size), debug=True
        )
        rsync = execnet.RSync(bundle_file, verbose=False)
        rsync.add_target(
            host.gateway, host.remote_repository + "/batou-bundle.git"
        )
        rsync.send()
        os.unlink(bundle_file)
        output.annotate("Unbundling changes", debug=True)
        host.rpc.git_unbundle_code()



import os
import shutil

import pytest


@pytest.fixture()
def sample_service(tmpdir):
    shutil.copytree(
        os.path.dirname(__file__) + "/fixture/sample_service",
        str(tmpdir / "sample_service"),
    )
    target = str(tmpdir / "sample_service")
    os.chdir(target)
    return target


@pytest.fixture(autouse=True)
def ensure_git_config(monkeypatch):
    monkeypatch.setitem(os.environ, "GIT_AUTHOR_EMAIL", "test@example.com")
    monkeypatch.setitem(os.environ, "GIT_AUTHOR_NAME", "Mr. U. Test")
    monkeypatch.setitem(os.environ, "GIT_COMMITTER_EMAIL", "test@example.com")
    monkeypatch.setitem(os.environ, "GIT_COMMITTER_NAME", "Mr. U. Test")



import os.path

import pytest

import batou.lib.git
from batou.utils import cmd


def _repos_path(root, name):
    repos_path = os.path.join(root.environment.workdir_base, name)
    cmd(
        "mkdir {dir}; cd {dir}; git init;"
        "git config user.name Jenkins;"
        "git config user.email jenkins@example.com;"
        "touch foo; git add .;"
        'git commit -am "foo"'.format(dir=repos_path)
    )
    return repos_path


@pytest.fixture(scope="function")
def repos_path(root, name="upstream"):
    return _repos_path(root, name)


@pytest.fixture(scope="function")
def repos_path2(root, name="upstream2"):
    return _repos_path(root, name)


@pytest.mark.slow
def test_runs_git_to_clone_repository(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master"
    )
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/clone/foo")
    )
    root.component.deploy()  # trigger verify


@pytest.mark.slow
def test_directly_after_clone_nothing_is_merged(root, repos_path):
    # When a Clone is confirgured with a branch, the general gesture for
    # updating is "git merge origin/branch", but directly after cloning, the
    # Clone is still on master, so this would try to merge the configured
    # branch into master, which is wrong.
    cmd(
        "cd {dir}; git checkout -b other; touch bar; echo qux > foo;"
        'git add .; git commit -am "other";'
        # Set up branches to be different, so we see that no merge takes place
        "git checkout master; "
        'echo one > foo; git add . ; git commit -am "foo master";'.format(
            dir=repos_path
        )
    )
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="other"
    )
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/clone/bar")
    )


@pytest.mark.slow
def test_setting_branch_updates_on_incoming_changes(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master"
    )
    root.component.deploy()
    cmd(
        'cd {dir}; touch bar; git add .; git commit -m "commit"'.format(
            dir=repos_path
        )
    )
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/clone/bar")
    )


@pytest.mark.slow
def test_setting_revision_updates_on_incoming_changes(root, repos_path):
    cmd(
        'cd {dir}; touch bar; git add .; git commit -m "commit2"'.format(
            dir=repos_path
        )
    )
    commit1, _ = cmd("cd {dir}; git rev-parse HEAD^".format(dir=repos_path))
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", revision=commit1
    )
    root.component.deploy()
    cmd(
        'cd {dir}; touch qux; git add .; git commit -m "commit3"'.format(
            dir=repos_path
        )
    )
    root.component.deploy()  # Our main assertion: Nothing breaks here
    assert not os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/clone/qux")
    )


@pytest.mark.slow
def test_branch_does_switch_branch(root, repos_path):
    cmd(
        "cd {dir}; touch bar; git add .; git checkout -b bar;"
        'git commit -m "commit branch"'.format(dir=repos_path)
    )
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="bar"
    )
    root.component.deploy()
    stdout, stderr = cmd(
        "cd {workdir}/clone; git rev-parse --abbrev-ref HEAD".format(
            workdir=root.workdir
        )
    )
    assert "bar" == stdout.strip()


@pytest.mark.slow
def test_tag_does_switch_tag(root, repos_path):
    cmd(
        """cd {dir}; git tag -a v1.0 -m "version 1.0" """.format(dir=repos_path)
    )
    cmd(
        "cd {dir}; touch bar; git add .;"
        'git commit -m "commit branch"'.format(dir=repos_path)
    )
    cmd(
        """cd {dir}; git tag -a v1.1 -m "version 1.1" """.format(dir=repos_path)
    )

    for tag in ("v1.0", "v1.1"):
        root.component += batou.lib.git.Clone(
            repos_path, target="clone", tag=tag
        )
        root.component.deploy()
        stdout, stderr = cmd(
            "cd {workdir}/clone; git describe --tags".format(
                workdir=root.workdir
            )
        )
        assert tag == stdout.strip()


@pytest.mark.slow
def test_has_changes_counts_changes_to_tracked_files(root, repos_path):
    clone = batou.lib.git.Clone(repos_path, target="clone", branch="master")
    root.component += clone
    root.component.deploy()
    assert not clone.has_changes()
    cmd("touch {}/clone/bar".format(root.workdir))
    cmd("cd {}/clone; git add bar".format(root.workdir))
    assert clone.has_changes()


@pytest.mark.slow
def test_has_changes_counts_untracked_files_as_changes(root, repos_path):
    clone = batou.lib.git.Clone(repos_path, target="clone", branch="master")
    root.component += clone
    root.component.deploy()
    assert not clone.has_changes()
    cmd("touch {}/clone/bar".format(root.workdir))
    assert clone.has_changes()


@pytest.mark.slow
def test_clean_clone_updates_on_incoming_changes(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master"
    )
    root.component.deploy()
    cmd(
        'cd {dir}; touch bar; git add .; git commit -m "commit"'.format(
            dir=repos_path
        )
    )
    root.component.deploy()
    assert os.path.isfile(root.component.map("clone/bar"))


@pytest.mark.slow
def test_no_clobber_changes_protected_on_update_with_incoming(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master"
    )
    root.component.deploy()
    cmd(
        'cd {dir}; touch bar; git add .; git commit -m "commit"'.format(
            dir=repos_path
        )
    )
    cmd("cd {dir}/clone; echo foobar >foo".format(dir=root.workdir))
    with pytest.raises(RuntimeError) as e:
        root.component.deploy()
    assert e.value.args[0] == "Refusing to clobber dirty work directory."
    with open(root.component.map("clone/foo")) as f:
        assert f.read() == "foobar\n"


@pytest.mark.slow
def test_no_clobber_changes_protected_on_update_without_incoming(
    root, repos_path
):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master"
    )
    root.component.deploy()
    cmd("cd {dir}/clone; echo foobar >foo".format(dir=root.workdir))
    with pytest.raises(RuntimeError) as e:
        root.component.deploy()
    assert e.value.args[0] == "Refusing to clobber dirty work directory."
    with open(root.component.map("clone/foo")) as f:
        assert f.read() == "foobar\n"


@pytest.mark.slow
def test_no_clobber_untracked_files_are_kept_on_update(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master"
    )
    root.component.deploy()
    cmd(
        "cd {dir}/clone; mkdir bar; echo foobar >bar/baz".format(
            dir=root.workdir
        )
    )
    with pytest.raises(RuntimeError) as e:
        root.component.deploy()
    assert e.value.args[0] == "Refusing to clobber dirty work directory."
    with open(root.component.map("clone/bar/baz")) as f:
        assert f.read() == "foobar\n"


@pytest.mark.slow
def test_clobber_changes_lost_on_update_with_incoming(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master", clobber=True
    )
    root.component.deploy()
    cmd(
        'cd {dir}; touch bar; git add .; git commit -m "commit"'.format(
            dir=repos_path
        )
    )
    cmd("cd {dir}/clone; echo foobar >foo".format(dir=root.workdir))
    root.component.deploy()
    assert os.path.exists(root.component.map("clone/bar"))
    with open(root.component.map("clone/foo")) as f:
        assert not f.read()


@pytest.mark.slow
def test_clobber_changes_lost_on_update_without_incoming(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master", clobber=True
    )
    root.component.deploy()
    cmd("cd {dir}/clone; echo foobar >foo".format(dir=root.workdir))
    root.component.deploy()
    with open(root.component.map("clone/foo")) as f:
        assert not f.read()


@pytest.mark.slow
def test_clobber_untracked_files_are_removed_on_update(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master", clobber=True
    )
    root.component.deploy()
    cmd(
        "cd {dir}/clone; mkdir bar; echo foobar >bar/baz".format(
            dir=root.workdir
        )
    )
    root.component.deploy()
    assert not os.path.exists(root.component.map("clone/bar/baz"))


@pytest.mark.slow
def test_clean_clone_vcs_update_false_leaves_changes_intact(root, repos_path):
    root.component += batou.lib.git.Clone(
        repos_path, target="clone", branch="master", vcs_update=False
    )
    root.component.deploy()
    cmd(
        "cd {dir}; echo foobar >foo; touch bar; git add .; "
        'git commit -m "commit"'.format(dir=repos_path)
    )
    cmd("cd {dir}/clone; echo asdf >foo".format(dir=root.workdir))
    root.component.deploy()
    with open(root.component.map("clone/foo")) as f:
        assert "asdf\n" == f.read()
    assert not os.path.exists(root.component.map("clone/bar"))


@pytest.mark.slow
def test_changed_remote_is_updated(root, repos_path, repos_path2):
    git = batou.lib.git.Clone(repos_path, target="clone", branch="master")
    root.component += git

    # Fresh, unrelated repo
    cmd(
        "cd {dir}; echo baz >bar; git add .;"
        'git commit -m "commit"'.format(dir=repos_path2)
    )

    root.component.deploy()
    assert not os.path.exists(root.component.map("clone/bar"))
    git.url = repos_path2
    root.component.deploy()
    assert os.path.exists(root.component.map("clone/bar"))


@pytest.mark.slow
def test_clone_into_workdir_works(root, repos_path, repos_path2):
    git = batou.lib.git.Clone(repos_path, branch="master")
    with open(root.component.map("asdf"), "w") as f:
        f.write("foo")
    root.component += git
    root.component.deploy()
    assert not os.path.exists(root.component.map("asdf"))


@pytest.mark.slow
def test_can_read_untracked_files_correctly(root, repos_path):
    fun_strings = [
        "this string has spaces in it",
        "this string has a ' in it",
        'this string has a " in it',
        '"this string has a \' and a " in it"',
        "this string has a \\ in it",
        "this string has a \\ and a ' in it",
        "this string has a \\ and a & in it",
    ]
    git = batou.lib.git.Clone(repos_path, branch="master")
    root.component += git
    root.component.deploy()
    for s in fun_strings:
        with open(root.component.map(s), "w") as f:
            f.write(s)
    with git.chdir(git.target):
        assert set(git.untracked_files()) == set(fun_strings)



import contextlib
import copy
import fcntl
import functools
import hashlib
import inspect
import itertools
import os
import re
import shlex
import socket
import subprocess
import sys
import time
from collections import defaultdict
from typing import Optional

import importlib_metadata

from batou import (
    DeploymentError,
    GetAddressInfoError,
    IPAddressConfigurationError,
    output,
)


class BagOfAttributes(dict):
    """Provide a dict-like object that can also
    be accessed using attributes.

    It's sometimes more convenient to write
    a.x instead of a['x']. However, namespaces may
    require being able to also use non-Python-identifier
    keys.

    """

    def __getattr__(self, key):
        return self[key]


def self_id():
    template = "batou/{version} ({python}, {system})"
    system = os.uname()
    system = " ".join([system[0], system[2], system[4]])
    version = importlib_metadata.version("batou")
    python = sys.implementation.name
    python += " {0}.{1}.{2}-{3}{4}".format(*sys.version_info)
    return template.format(**locals())


class MultiFile(object):
    def __init__(self, files):
        self.files = files

    def write(self, value):
        for file in self.files:
            file.write(value)

    def flush(self):
        for file in self.files:
            file.flush()


@contextlib.contextmanager
def locked(filename, exit_on_failure=False):
    # XXX can we make this not leave files around?
    with open(filename, "a+") as lockfile:
        try:
            fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            print("Could not acquire lock {}".format(filename), file=sys.stderr)
            if exit_on_failure:
                print(
                    "Another instance of batou is currently running and locked the lockfile.",
                    file=sys.stderr,
                )
                print("Exiting.", file=sys.stderr)
                sys.exit(1)
            raise RuntimeError(
                'cannot create lock "%s": more than one instance running '
                "concurrently?" % lockfile,
                lockfile,
            )
        # publishing the process id comes handy for debugging
        lockfile.seek(0)
        lockfile.truncate()
        print(os.getpid(), file=lockfile)
        lockfile.flush()
        yield
        lockfile.seek(0)
        lockfile.truncate()


def flatten(list_of_lists):
    return list(itertools.chain.from_iterable(list_of_lists))


def notify_send(title, description):
    subprocess.call(["notify-send", title, description])


def notify_macosx(title, description):
    subprocess.call(
        [
            "osascript",
            "-e",
            'display notification "{}" with title "{}"'.format(
                description, title
            ),
        ]
    )


def notify_none(title, description):
    pass


try:
    subprocess.check_output(["which", "osascript"], stderr=subprocess.STDOUT)
    notify = notify_macosx
except (subprocess.CalledProcessError, OSError):
    try:
        subprocess.check_output(
            ["which", "notify-send"], stderr=subprocess.STDOUT
        )
        notify = notify_send
    except (subprocess.CalledProcessError, OSError):
        notify = notify_none

resolve_override = {}
resolve_v6_override = {}


def resolve(host, port=0, resolve_override=resolve_override):
    if host in resolve_override:
        address = resolve_override[host]
        output.annotate(
            "resolved (v4) `{}` to {} (override)".format(host, address),
            debug=True,
        )
    else:
        output.annotate("resolving (v4) `{}`".format(host), debug=True)
        responses = socket.getaddrinfo(host, int(port), socket.AF_INET)
        output.annotate(
            "resolved (v4) `{}` to {}".format(host, responses), debug=True
        )
        address = responses[0][4][0]
        output.annotate(
            "selected (v4) {}, {}".format(host, address), debug=True
        )
    return address


def resolve_v6(host, port=0, resolve_override=resolve_v6_override):
    if host in resolve_override:
        address = resolve_override[host]
        output.annotate(
            "resolved (v6) `{}` to {} (override)".format(host, address),
            debug=True,
        )
    else:
        output.annotate("resolving (v6) `{}`".format(host), debug=True)
        responses = socket.getaddrinfo(host, int(port), socket.AF_INET6)
        output.annotate(
            "resolved (v6) `{}` to {}".format(host, responses), debug=True
        )
        address = None
        for _, _, _, _, sockaddr in responses:
            addr = sockaddr[0]
            if addr.startswith("fe80:"):
                continue
            address = addr
            break
        if not address:
            raise ValueError("No valid address found for `{}`.".format(host))
        output.annotate(
            "selected (v6) {}, {}".format(host, address), debug=True
        )
    return address


ADDR_DEFAULT = object()  # sentinel


@functools.total_ordering
class Address(object):
    """An internet service address that can be listened and connected to.

    The constructor address is expected to be the address that can be
    connected to. The listen address will be computed automatically.

    .. code-block:: pycon

        >>> x = Address('localhost', 80)
        >>> str(x.connect)
        'localhost:80'
        >>> str(x.listen)
        '127.0.0.1:80'

    You can specify which IP versions are expected to be resolved for listen
    addresses in three ways with the require_v4/require_v6 flags:

    * False -> this version must not be used
    * True -> this version must be resolved properly
    * 'optional' -> the listen/listen_v6 attribute will contain None if it does
       not resolve.

    """

    #: The connect address as it should be used when configuring clients.
    #: This is a :py:class:`batou.utils.NetLoc` object.
    connect: "NetLoc"

    # The dance with the ADDR_DEFAULT is to allow overriding the class-based
    # defaults on an environment level.
    require_v4 = True
    require_v6 = False

    def __init__(
        self,
        connect_address,
        port=None,
        require_v4=ADDR_DEFAULT,
        require_v6=ADDR_DEFAULT,
    ):
        if require_v4 is not ADDR_DEFAULT:
            self.require_v4 = require_v4
        if require_v6 is not ADDR_DEFAULT:
            self.require_v6 = require_v6

        if not self.require_v4 and not self.require_v6:
            raise ValueError(
                "At least one of `require_v4` or `require_v6` is required. "
                "None were selected."
            )
        if ":" in connect_address:
            connect, port = connect_address.split(":")
        else:
            connect = connect_address
        if port is None:
            raise ValueError("Need port for service address.")
        self.connect = NetLoc(connect, port)

    def __lt__(self, other):
        if isinstance(other, Address):
            return str(self) < str(other)
        return NotImplemented

    def __eq__(self, other):
        if isinstance(other, Address):
            return str(self) == str(other)
        return NotImplemented

    def __str__(self):
        return str(self.connect)

    @property
    def listen(self):
        """The IPv4 listen (or bind) address as it should be used when
        configuring servers. This is a :py:class:`batou.utils.NetLoc`
        object. It raises an :py:class:`batou.IPAddressConfigurationError`
        if used unconfigured.
        """
        if not self.require_v4:
            raise IPAddressConfigurationError.from_context(self, 4)

        assert self.connect.port is not None
        if not hasattr(self, "_listen_v4"):
            try:
                address = resolve(self.connect.host, self.connect.port)
                self.listen = NetLoc(address, self.connect.port)
            except Exception as e:
                if self.require_v4 == "optional":
                    self.listen = None
                else:
                    if isinstance(e, OSError):
                        raise GetAddressInfoError.from_context(
                            f"{self.connect.host}:{self.connect.port}", str(e)
                        )
                    else:
                        raise

        return self._listen_v4

    @listen.setter
    def listen(self, value):
        self._listen_v4 = value

    @property
    def listen_v6(self):
        """The IPv6 listen (or bind) address as it should be used when
        configuring servers. This is a :py:class:`batou.utils.NetLoc`
        object. It raises an :py:class:`batou.IPAddressConfigurationError`
        if used unconfigured.
        """
        if not self.require_v6:
            raise IPAddressConfigurationError.from_context(self, 6)

        assert self.connect.port is not None
        if not hasattr(self, "_listen_v6"):
            try:
                address = resolve_v6(self.connect.host, self.connect.port)
                self.listen_v6 = NetLoc(address, self.connect.port)
            except Exception as e:
                if self.require_v6 == "optional":
                    self.listen_v6 = None
                else:
                    if isinstance(e, OSError):
                        raise GetAddressInfoError.from_context(
                            f"{self.connect.host}:{self.connect.port}", str(e)
                        )
                    else:
                        raise

        return self._listen_v6

    @listen_v6.setter
    def listen_v6(self, value):
        self._listen_v6 = value


@functools.total_ordering
class NetLoc(object):
    """A network location specified by host and port.

    Network locations can automatically render an appropriate string
    representation:

    .. code-block:: pycon

        >>> x = NetLoc('127.0.0.1')
        >>> x.host
        '127.0.0.1'
        >>> x.port
        None
        >>> str(x)
        '127.0.0.1'

        >>> y = NetLoc('127.0.0.1', 80)
        >>> str(y)
        '127.0.0.1:80'

    """

    #: The host part of this network location. Can be a hostname or IP address.
    host: str
    #: The port of this network location. Can be ``None`` or an integer.
    port = None

    def __init__(self, host, port=None):
        self.host = host
        self.port = port

    def __str__(self):
        if self.port:
            if ":" in self.host:  # ipv6
                fmt = "[{self.host}]:{self.port}"
            else:
                fmt = "{self.host}:{self.port}"
        else:
            fmt = "{self.host}"
        return fmt.format(self=self)

    def __repr__(self):
        return "<NetLoc `{}`>".format(self)

    # These are not "correct" comparisons from a networking viewpoint.
    # However, they are useful to provide a predictable ordering to
    # avoid unnecessary config changes. Also, the values do not have
    # to be IP addresses but can be hostnames as well.
    def __lt__(self, other):
        return str(self) < other

    def __eq__(self, other):
        return str(self) == other


def revert_graph(graph):
    graph = ensure_graph_data(graph)
    reverse_graph = defaultdict(set)
    for node, dependencies in list(graph.items()):
        # Ensure all nodes will exist
        reverse_graph[node]
        for dependency in dependencies:
            reverse_graph[dependency].add(node)
    return reverse_graph


def ensure_graph_data(graph):
    # Ensure that all nodes exist as keys even if they don't have outgoing
    # relations.
    for node, relations in list(graph.items()):
        for relation in relations:
            if relation not in graph:
                graph[relation] = set()
    return graph


class CycleError(ValueError):
    def __str__(self):
        message = []
        components = list(self.args[0].items())
        components.sort(key=lambda x: x[0].name)
        for component, subs in components:
            message.append(component.name + " depends on")
            for sub in subs:
                message.append("        " + sub.name)
        return "\n".join(message)


def remove_nodes_without_outgoing_edges(graph):
    for node, dependencies in list(graph.items()):
        if not dependencies:
            del graph[node]


def topological_sort(graph):
    # Take a directed graph and provide a topological sort of all nodes.
    #
    # The graph is given as
    #
    # {node: [dependency, dependency], ...}
    #
    # If the graph has cycles a CycleError will be raised.

    graph = ensure_graph_data(graph)
    sorted = []
    reverse_graph = revert_graph(graph)
    roots = [
        node for node, incoming in list(reverse_graph.items()) if not incoming
    ]
    while roots:
        root = roots.pop()
        sorted.append(root)
        for node in list(graph[root]):
            graph[root].remove(node)
            reverse_graph[node].remove(root)
            if not reverse_graph[node]:
                roots.append(node)
    if any(graph.values()):
        # Simplify the graph a bit to make it easier to spot the cycle.
        remove_nodes_without_outgoing_edges(graph)
        raise CycleError(dict(graph))
    return sorted


class CmdExecutionError(DeploymentError, RuntimeError):
    def __init__(self, cmd, returncode, stdout, stderr):
        self.cmd = cmd
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.args = (cmd, returncode, stdout, stderr)

    def report(self):
        output.error(self.cmd)
        output.tabular("Return code", str(self.returncode), red=True)
        output.line("STDOUT", red=True)
        output.annotate(self.stdout)
        output.line("STDERR", red=True)
        output.annotate(self.stderr)

    def __str__(self):
        return f"Command {self.cmd} failed with return code {self.returncode}\n\nSTDOUT:\n{self.stdout}\n\nSTDERR:\n{self.stderr}"


def cmd(
    cmd,
    silent=False,
    ignore_returncode=False,
    communicate=True,
    env=None,
    acceptable_returncodes=[0],
    encoding="utf-8",
):
    if not isinstance(cmd, str):
        # We use `shell=True`, so the command needs to be a single string and
        # we need to pay attention to shell quoting.
        quoted_args = []
        for arg in cmd:
            arg = arg.replace("'", "\\'")
            if " " in arg:
                arg = "'{}'".format(arg)
            quoted_args.append(arg)
        cmd = " ".join(quoted_args)
    if env is not None:
        add_to_env = env
        env = os.environ.copy()
        env.update(add_to_env)
    output.annotate("cmd: {}".format(cmd), debug=True)
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        shell=True,
        env=env,
    )
    if not communicate:
        # XXX See #12550
        return process
    stdout, stderr = process.communicate()
    if encoding is not None:
        stdout = stdout.decode(encoding, errors="replace")
        stderr = stderr.decode(encoding, errors="replace")
    if process.returncode not in acceptable_returncodes:
        if not ignore_returncode:
            raise CmdExecutionError(cmd, process.returncode, stdout, stderr)
    return stdout, stderr


def get_output(command, default=None):
    stdout, stderr = cmd(command, ignore_returncode=True)  # type: ignore
    return stdout or default


class Timer(object):
    def __init__(self, tag=None):
        self.durations = defaultdict(float)  # returns 0.0 for missing keys
        self.tag = tag

    class TimerContext(object):
        def __init__(self, timer, note):
            self.timer = timer
            self.note = note

        def __enter__(self):
            self.start = time.time()

        def __exit__(self, exc_type, exc_value, traceback):
            delta = time.time() - self.start
            self.timer.durations[self.note] += delta
            output.annotate(
                f"Timer {self.timer.tag}: {self.note} took {delta:.2f} seconds",
                debug=True,
            )
            self.timer.durations[self.note] += time.time() - self.start

    def step(self, note):
        if note == "total":
            raise ValueError("Cannot use 'total' as a step name")
        return self.TimerContext(self, note)

    def above_threshold(self, **thresholds):
        """
        Return true if any of the steps took longer than the given threshold.
        "total" is a special step that is the sum of all other steps.
        """

        total = sum(self.durations.values())
        for note, duration in self.durations.items():
            if note in thresholds and duration > thresholds[note]:
                return True
        if "total" in thresholds and total > thresholds["total"]:
            return True
        return False

    def humanize(self, *steps):
        """
        Given a list of steps, return a string with the time taken for each
        step.
        """
        total = sum(self.durations.values())
        durations = self.durations.copy()
        durations["total"] = total

        return ", ".join(
            f"{note}={format_duration(durations.get(note))}" for note in steps
        )


def hash(path, function="sha_512"):
    h = getattr(hashlib, function)()
    with open(path, "rb") as f:
        chunk = f.read(64 * 1024)
        while chunk:
            h.update(chunk)
            chunk = f.read(64 * 1024)
    return h.hexdigest()


def call_with_optional_args(func, **kw):
    """Provide a way to perform backwards-compatible call,
    passing only arguments that the function actually expects.
    """
    call_kw = {}
    verify_args = inspect.signature(func)
    for name, parameter in verify_args.parameters.items():
        if name in kw:
            call_kw[name] = kw[name]
        if parameter.kind == inspect.Parameter.VAR_KEYWORD:
            call_kw = kw
            break
    return func(**call_kw)


def dict_merge(a, b):
    """recursively merges dict's. not just simple a['key'] = b['key'], if
    both a and b have a key who's value is a dict then dict_merge is called
    on both values and the result stored in the returned dictionary.
    https://www.xormedia.com/recursively-merge-dictionaries-in-python/
    """
    if not isinstance(b, dict):
        return b
    result = copy.deepcopy(a)
    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(result[k], v)
        elif k in result and isinstance(result[k], list):
            result[k] = result[k][:]
            result[k].extend(v)
        else:
            result[k] = copy.deepcopy(v)
    return result


def format_duration(duration: Optional[float]) -> str:
    """
    Formats a duration (in seconds) into a human readable string.

    Examples:
    ```
    format_duration(1.23124) == "1.23s"
    format_duration(61) == "1m1s"
    format_duration(None) == "NaN"
    ```
    """

    if duration is None:
        return "NaN"

    minutes, seconds = divmod(duration, 60)

    if minutes:
        output = "{}m{}s".format(int(minutes), int(seconds))
    else:
        output = "{:.2f}s".format(seconds)

    return output


VARNAME_PATTERN = re.compile("^[a-zA-Z_][a-zA-Z_0-9]*$")


def export_environment_variables(environ):
    """Turn this dict into shell environment variable exports.

    This creates a snippet that can be embedded into a shell script to access
    all entries of this dict as environment variables.

    Ensures proper quoting.
    """
    # Verify all keys are valid as shell variable nameså
    for k in environ:
        if not VARNAME_PATTERN.match(k):
            raise ValueError(f"`{k}` is not a valid shell variable name")
    return "\n".join(
        sorted(f"export {k}={shlex.quote(v)}" for k, v in environ.items())
    )



import os
import tempfile

from batou import UpdateNeeded
from batou.component import Component

# Pass `USE_SUDO` as `admin_password` to use sudo for authorisation, like we
# have on FC platform.
USE_SUDO = object()


class Command(Component):

    namevar = "statement"
    admin_password = ""
    admin_user = "root"
    hostname = None
    port = None

    db = "mysql"

    # Unless allows you to specify a query that generates output if the
    # target state has been reached already.
    unless = ""

    def configure(self):
        self.statement = self.expand(self.statement)
        self.unless = self.expand(self.unless)

    def _mysql(self, cmd):
        _, self.tmp = tempfile.mkstemp(suffix="sql")
        with open(self.tmp, "w") as f:
            f.write(cmd + "\n")

        command = []

        if self.admin_password is USE_SUDO:
            command.append("sudo -u mysql")

        command.append("mysql -Bs")

        if self.admin_password is not USE_SUDO:
            command.append("-u{{component.admin_user}}")
            command.append("-p{{component.admin_password}}")

        if self.hostname:
            command.append("-h {{component.hostname}}")
        if self.port:
            command.append("-P {{component.port}}")
        command.append("{{component.db}} < {{component.tmp}}")

        out, err = self.cmd(self.expand(" ".join(command)))
        os.remove(self.tmp)
        return out, err

    def verify(self):
        if not self.unless:
            raise UpdateNeeded()
        out, err = self._mysql(self.unless)
        if not out.strip():
            raise UpdateNeeded()

    def update(self):
        self._mysql(self.statement)

    @property
    def namevar_for_breadcrumb(self):
        words = self.statement.split()
        if words:
            return " ".join(words[:2]).upper()
        return "--missing--"


class Database(Component):

    namevar = "database"
    charset = "UTF8"
    base_import_file = None
    admin_password = None

    def configure(self):
        create_db = self.expand(
            """\
CREATE DATABASE IF NOT EXISTS
    {{component.database}}
    DEFAULT CHARACTER SET = '{{component.charset}}';
"""
        )
        self += Command(create_db, admin_password=self.admin_password)

        if self.base_import_file:
            self += Command(
                self.expand("\\. {{component.base_import_file}}\n"),
                db=self.database,
                unless=self.expand("show tables"),
                admin_password=self.admin_password,
            )


class User(Component):

    namevar = "user"
    password = None
    allow_from_hostname = "localhost"
    admin_password = None

    SET_PASSWORD_QUERY = """
SET PASSWORD FOR
    '{{component.user}}'@'{{component.allow_from_hostname}}' =
     '{{component.password}}';
"""

    def configure(self):

        create = self.expand(
            """\
CREATE USER '{{component.user}}'@'{{component.allow_from_hostname}}';
"""
        )
        create_unless = self.expand(
            """\
SELECT *
FROM user
WHERE
    User = '{{component.user}}'
    AND
    Host = '{{component.allow_from_hostname}}';
"""
        )
        self += Command(
            create, unless=create_unless, admin_password=self.admin_password
        )
        self += Command(
            self.expand(self.SET_PASSWORD_QUERY),
            admin_password=self.admin_password,
        )


class Grant(Command):

    namevar = "grant_db"
    user = ""
    allow_from_hostname = "localhost"
    statement = """\
GRANT ALL
    ON {{component.grant_db}}.*
    TO '{{component.user}}'@'{{component.allow_from_hostname}}';
"""



import hashlib
import os.path

from batou.component import Component
from batou.lib.file import File, Symlink, ensure_path_nonexistent


class VirtualEnv(Component):

    namevar = "python_version"
    pip_version = None

    def verify(self):
        assert os.path.exists(self.parent.env_ready)
        # XXX Check if python is executable

    def update(self):
        ensure_path_nonexistent(self.parent.env_dir)
        if "__PYVENV_LAUNCHER__" in os.environ:
            del os.environ["__PYVENV_LAUNCHER__"]
        self.cmd(
            "python{{component.python_version}} -m venv "
            "{{component.parent.env_dir}}"
        )
        if self.pip_version:
            self.cmd(
                "{{component.parent.env_dir}}/bin/pip "
                f"install pip=={self.pip_version}"
            )
        else:
            self.cmd(
                "{{component.parent.env_dir}}/bin/python -m pip "
                "install --upgrade pip"
            )


class LockedRequirements(Component):
    def verify(self):
        assert os.path.exists(self.parent.env_ready)

    def update(self):
        self.python = os.path.join(self.parent.env_dir, "bin", "python")
        self.cmd(
            "{{component.python}} -m pip install --no-deps "
            "-r {{component.parent.env_dir}}/requirements.lock"
        )
        self.cmd("{{component.python}} -m pip check")


class CleanupUnused(Component):

    cleanup = ()

    def verify(self):
        if not os.path.exists(".appenv/"):
            return
        protected = set(
            [self.parent.env_hash, self.parent.last_env_hash, "current"]
        )
        self.cleanup = set(os.listdir(".appenv/")) - protected
        assert not self.cleanup

    def update(self):
        for path in self.cleanup:
            ensure_path_nonexistent(os.path.join(".appenv", path))


class AppEnv(Component):
    """Manage a Python-package based application installation.

    This uses virtualenv and pip with a "lockfile" style requirements.txt
    file.

    Place a file name "requirements.lock" in the defdir and specify the version
    of Python 3 to be used.

    Will automatically rebuild based on hashing of the AppEnv code,
    the requirements file and the chosen version of Python.

    The result is a light-weight 'pseudo' virtualenv with a bin/ directory
    symlinked to the real virtualenv used under the hood.

    """

    namevar = "python_version"
    pip_version = None

    def configure(self):
        with open("requirements.lock", "r") as f:
            lockfile = f.read()
        with open(__file__, "r") as f:
            hash_content = lockfile + self.python_version + f.read()
        hash_content = hash_content.encode("utf-8")
        self.env_hash = hashlib.new("sha256", hash_content).hexdigest()[:8]
        self.env_dir = os.path.join(".appenv", self.env_hash)
        self.env_ready = os.path.join(self.env_dir, "appenv.ready")

        self += VirtualEnv(self.python_version, pip_version=self.pip_version)
        self += File(
            os.path.join(self.env_dir, "requirements.lock"), content=lockfile
        )
        self += LockedRequirements()

        # If we got here, then we can place the ready marker
        self += File(
            os.path.join(self.env_dir, "appenv.ready"),
            content="Ready or not, here I come, you can't hide\n",
        )

        # Save current to skip it in cleanup
        self.last_env_hash = None
        if os.path.exists(os.path.join(self.workdir, ".appenv/current")):
            self.last_env_hash = os.path.basename(
                os.readlink(os.path.join(self.workdir, ".appenv/current"))
            )

        # Shim
        self += Symlink(".appenv/current", source=self.env_dir)
        for p in ["bin", "lib", "pyvenv.cfg"]:
            self += Symlink(p, source=".appenv/current/" + p)

        self += CleanupUnused()

    @property
    def namevar_for_breadcrumb(self):
        return self.env_hash[:8]



import os
import os.path

import pytest

import batou
from batou import MissingEnvironment
from batou.component import Component
from batou.environment import Environment


def test_parse_nonexisting_environment_raises_error(tmpdir):
    environment = Environment("test", basedir=str(tmpdir))
    with pytest.raises(MissingEnvironment):
        environment.load()


@pytest.fixture
def env():
    environment = Environment(
        "dev", basedir=os.path.dirname(__file__) + "/fixture/basic_service"
    )
    environment.load()
    return environment


@pytest.fixture
def env_prod():
    environment = Environment(
        "production",
        basedir=os.path.dirname(__file__) + "/fixture/basic_service",
    )
    environment.load()
    return environment


def test_service_options_are_set(env):
    assert env.base_dir.endswith("/fixture/basic_service")


def test_components_are_fully_loaded(env):
    assert sorted(env.components) == ["zeo", "zope"]


def test_production_environment_is_loaded(env_prod):
    assert env_prod.service_user == "alice"
    assert env_prod.branch == "production"
    assert env_prod.host_domain == "example.com"

    assert sorted(env_prod.hosts) == [
        "host1",
        "host2",
        "host3",
    ]

    host1 = env_prod.hosts["host1"]
    assert host1.name == "host1"
    assert host1.fqdn == "host1.example.com"
    assert host1.service_user == "alice"
    host1_components = [
        x.name for x in env_prod.root_components if x.host is host1
    ]
    assert host1_components == ["zope"]


def test_dev_environment_is_loaded(env):
    assert env.branch == "default"
    assert env.host_domain is None
    assert set(env.hosts.keys()) == set(["localhost", "host2"])

    localhost = env.hosts["localhost"]
    assert localhost.name == "localhost"
    assert localhost.fqdn == "localhost"
    root_components = set(
        [x.name for x in env.root_components if x.host is localhost]
    )
    assert root_components == set(["zeo", "zope"])

    zeo = env.get_root("zeo", env.hosts["localhost"])
    zeo.prepare()
    assert isinstance(zeo.component, Component)
    assert zeo.name == "zeo"
    assert zeo.component.port == 9001


def test_component_has_features_set(env):
    env.configure()
    # The localhost ZEO doesn't specify features and thus gets all set
    zeo1 = env.get_root("zeo", env.hosts["localhost"]).component
    assert zeo1.features == ["test", "test2"]

    # ZEO on host2 has selected only 'test'
    zeo2 = env.get_root("zeo", env.hosts["host2"]).component
    assert zeo2.features == ["test"]


def test_load_environment_with_overrides(env):
    env.overrides["zeo"] = {"port": "9002"}
    env.configure()
    zeo = env.get_root("zeo", env.hosts["localhost"]).component
    assert zeo.port == 9002


def test_config_exceptions_orderable(env):
    env.configure()
    c = env.get_root("zeo", env.hosts["localhost"]).component

    import sys

    try:
        raise ValueError("Test")
    except Exception:
        exc_type, ex, tb = sys.exc_info()

    exceptions = [
        batou.ConfigurationError.from_context("test", None),
        batou.ConfigurationError.from_context("test", c),
        batou.ConversionError.from_context(c, "key", 123, int, "invalid int"),
        batou.MissingOverrideAttributes.from_context(c, ["sadf"]),
        batou.DuplicateComponent.from_context(
            env.components["zeo"], env.components["zeo"]
        ),
        batou.UnknownComponentConfigurationError.from_context(c.root, ex, tb),
        batou.UnusedResources.from_context({"asdf": {c.root: 1}}),
        batou.UnsatisfiedResources.from_context({("asdf", None): [c.root]}),
        batou.MissingEnvironment.from_context(env),
        batou.MissingComponent.from_context("asdf", "localhost"),
        batou.SuperfluousSection.from_context("asdf"),
        batou.SuperfluousComponentSection.from_context("asdf"),
        batou.SuperfluousSecretsSection.from_context("asdf"),
        batou.CycleErrorDetected.from_context(ValueError()),
        batou.NonConvergingWorkingSet.from_context([c.root]),
        batou.DeploymentError(),
        batou.DuplicateHostMapping.from_context("host", "map1", "map2"),
        batou.RepositoryDifferentError.from_context("asdf", "bsdf"),
        batou.DuplicateHostError.from_context("localhost"),
        batou.InvalidIPAddressError.from_context("asdf"),
    ]

    # Ensure all exceptions can be compared
    for x in exceptions:
        for y in exceptions:
            x.sort_key < y.sort_key


def test_remote_pdb_config():
    """There are some environment variables set with default values."""
    assert os.environ["PYTHONBREAKPOINT"] == "remote_pdb.set_trace"
    assert os.environ["REMOTE_PDB_HOST"] == "127.0.0.1"
    assert os.environ["REMOTE_PDB_PORT"] == "4444"



import contextlib
import os

from batou.component import Component
from batou.lib.file import File, Presence
from batou.lib.python import Package, VirtualEnv


class Buildout(Component):

    timeout = None
    use_default = True
    config = None
    additional_config = ()
    config_file_name = "buildout.cfg"

    python = None
    executable = None
    distribute = None
    setuptools = None
    wheel = None
    pip = None
    version = None

    build_env = {}  # XXX not frozen. :/

    def configure(self):
        if self.timeout is None:
            self.timeout = self.environment.timeout
        if self.use_default and not self.config:
            # We expect that your definition directory has a buildout.cfg
            self.config = File("buildout.cfg", template_context=self.parent)
        if isinstance(self.config, File):
            self.config_file_name = self.config.path
        if isinstance(self.config, Component):
            self.config = [self.config]
        if not self.config:
            self.config = []
        self.config.extend(self.additional_config)
        for component in self.config:
            self += component

        venv = VirtualEnv(self.python, executable=self.executable)
        self += venv

        if not (self.distribute or self.setuptools):
            raise ValueError(
                "Either setuptools or distribute version must be specified"
            )

        if self.distribute:
            venv += Package(
                "distribute",
                version=self.distribute,
                check_package_is_module=False,
            )

        if self.setuptools:
            venv += Package("setuptools", version=self.setuptools)

        if self.wheel:
            venv += Package("wheel", version=self.wheel)

        if self.pip:
            venv += Package("pip", version=self.pip)

        # Install without dependencies (that's just setuptools anyway), since
        # that could cause pip to pull in the latest version of setuptools,
        # regardless of the version we wanted to be installed above.
        venv += Package("zc.buildout", version=self.version, dependencies=False)

    def verify(self):
        self.assert_file_is_current("bin/buildout")
        # XXX we can't be sure that all config objects are files!
        installed = Presence(".installed.cfg")
        self |= installed
        installed.assert_component_is_current(
            [Presence("bin/buildout")] + self.config
        )
        self.assert_file_is_current(
            ".batou.buildout.success", [".installed.cfg"]
        )
        self.assert_no_subcomponent_changes()

    def update(self):
        with safe_environment(self.build_env):
            self.cmd(
                'bin/buildout -t {} -c "{}"'.format(
                    self.timeout, self.config_file_name
                )
            )
            self.touch(".batou.buildout.success")


@contextlib.contextmanager
def safe_environment(environment):
    old_env = os.environ.copy()
    for key, value in list(environment.items()):
        old_env.setdefault(key, "")
        environment[key] = value.format(**old_env)
    try:
        os.environ.update(environment)
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_env)



from pathlib import Path


def migrate(output):
    """Move secrets to environments."""
    output(
        "Environments have separate directories now",
        """

Each environment has been turned into a separate directory containing all the
configuration files:

* unencrypted configuration is now stored in
  `environments/<env>/environment.cfg`
* encrypted configuration is now stored in `environments/<env>/secrets.cfg`
* additional encrypted files are stored in `environments/<env>/secret-*`

The old `secrets` directory is now gone.

All data was migrated automatically for you, but you will need to commit those
changes manually.

        """,
        "automatic",
    )
    environments = _get_environment_names()
    # sort environments by length to ensure that we migrate the most specific
    # environments first
    environments.sort(key=len, reverse=True)
    for name in environments:
        _migrate_environment(name)


def _get_environment_names():
    return [
        x.stem
        for x in (Path.cwd() / "environments").iterdir()
        if x.suffix == ".cfg"
    ]


def _migrate_environment(name: str) -> None:
    cwd = Path.cwd()
    environments = cwd / "environments"
    (environments / name).mkdir(exist_ok=True)
    (environments / f"{name}.cfg").rename(
        environments / name / "environment.cfg"
    )
    secrets_file = cwd / "secrets" / f"{name}.cfg"
    if secrets_file.exists():
        secrets_file.rename(environments / name / "secrets.cfg")
    try:
        for secrets_file in (cwd / "secrets").glob(f"{name}-*"):
            secrets_file.rename(
                environments
                / name
                / secrets_file.name.replace(f"{name}-", "secret-", 1)
            )
    except FileNotFoundError:
        pass



import os
import socket
import tempfile
import threading
import unittest
from io import StringIO

import mock
import pytest

import batou
from batou.utils import (
    Address,
    CmdExecutionError,
    MultiFile,
    NetLoc,
    call_with_optional_args,
    cmd,
    flatten,
    format_duration,
    hash,
    locked,
    notify,
    remove_nodes_without_outgoing_edges,
    resolve,
    resolve_v6,
    revert_graph,
    topological_sort,
)

RESOLVER_ERRORS = [
    "[Errno -2] Name or service not known",  # Linux
    "[Errno 8] nodename nor servname provided, or not known",  # MacOS
]


@mock.patch("socket.getaddrinfo")
def test_host_without_port_resolves(ghbn):
    ghbn.return_value = [(None, None, None, None, ("127.0.0.1", 0, None, None))]
    assert resolve("localhost") == "127.0.0.1"


@mock.patch("socket.getaddrinfo", side_effect=socket.gaierror("lookup failed"))
def test_resolve_v4_socket_error_returns_none(ghbn):
    with pytest.raises(socket.gaierror) as f:
        resolve("localhost", 80)
    assert "lookup failed" == str(f.value)


@mock.patch("socket.getaddrinfo", side_effect=socket.gaierror("lookup failed"))
def test_resolve_v6_raises_on_socket_error(gai):
    with pytest.raises(socket.gaierror) as f:
        resolve_v6("localhost", 22)
    assert "lookup failed" == str(f.value)


def test_resolve_override():
    ov = {"foo.example.com": "1.2.3.4"}
    assert "1.2.3.4" == resolve("foo.example.com", 80, resolve_override=ov)


def test_resolve_v6_override():
    ov = {"foo.example.com": "::27"}
    assert "::27" == resolve_v6("foo.example.com", 80, resolve_override=ov)


def test_resolve_v6_does_not_return_link_local_addresses(output, monkeypatch):
    output.enable_debug = True

    def link_local_addrinfo_1(*args, **kw):
        return [
            (
                None,
                None,
                None,
                None,
                ("fe80::feaa:14ff:fe8f:94ba", 80, None, None),
            )
        ]

    monkeypatch.setattr(socket, "getaddrinfo", link_local_addrinfo_1)

    with pytest.raises(ValueError) as err:
        resolve_v6("foo.example.com", 80)
    assert "No valid address found for `foo.example.com`." == str(err.value)

    def link_local_addrinfo_2(*args, **kw):
        return [
            (
                None,
                None,
                None,
                None,
                ("fe80::feaa:14ff:fe8f:94ba", 80, None, None),
            ),
            (None, None, None, None, ("2a02::1", 80, None, None)),
        ]

    monkeypatch.setattr(socket, "getaddrinfo", link_local_addrinfo_2)

    output.backend.output = ""
    assert resolve_v6("foo.example.com", 80) == "2a02::1"

    assert (
        """\
resolving (v6) `foo.example.com`
resolved (v6) `foo.example.com` to [(None, None, None, None, ('fe80::feaa:14ff:fe8f:94ba', 80, None, None)), (None, None, None, None, ('2a02::1', 80, None, None))]
selected (v6) foo.example.com, 2a02::1
"""
        == output.backend.output
    )  # noqa: E501 line too long


def test_address_without_implicit_or_explicit_port_fails():
    with pytest.raises(ValueError):
        Address("localhost")
    Address("localhost:8080")
    Address("localhost", 8080)


def test_address_resolves_listen_address():
    address = Address("localhost:8080")
    assert "127.0.0.1:8080" == str(address.listen)
    assert "localhost:8080" == str(address.connect)


def test_address_netloc_attributes():
    address = Address("localhost:8080")
    assert "127.0.0.1" == address.listen.host
    assert "8080" == address.listen.port
    assert "localhost" == address.connect.host
    assert "8080" == address.connect.port
    assert address.require_v4


def test_netloc_ordering():
    address1 = NetLoc("127.0.0.5")
    address2 = NetLoc("127.0.0.5", "80")
    address3 = NetLoc("asdf")
    address4 = NetLoc("asdf", "90")
    sorted_list = sorted([address3, address2, address4, address1])
    assert [address1, address2, address3, address4] == sorted_list


def test_netloc_repr():
    assert "<NetLoc `127.0.0.5:80`>" == repr(NetLoc("127.0.0.5", "80"))
    assert "<NetLoc `127.0.0.5`>" == repr(NetLoc("127.0.0.5"))


def test_address_ordering():
    address1 = Address("127.0.0.5:8080")
    address2 = Address("localhost:8080")
    address3 = Address("127.10.1.5:8080")
    address4 = Address("127.122.122.133:8080")
    sorted_list = sorted([address1, address2, address3, address4])
    assert [address1, address3, address4, address2] == sorted_list


def test_address_neither_v4_v6_invalid():
    with pytest.raises(ValueError) as f:
        Address("asdf", require_v4=False, require_v6=False)
    assert (
        "At least one of `require_v4` or `require_v6` is required. "
        "None were selected." == str(f.value)
    )


def test_address_v6_only(monkeypatch):
    hostname = "v6only.example.com"

    from batou.utils import resolve_v6_override

    monkeypatch.setitem(resolve_v6_override, hostname, "::346")

    with pytest.raises(batou.GetAddressInfoError) as f:
        a = Address(hostname, 1234)
        a.listen

    assert any(x in str(f.value) for x in RESOLVER_ERRORS)

    address = Address(hostname, 1234, require_v4=False, require_v6=True)
    assert address.listen_v6.host == "::346"
    assert address.require_v6


def test_address_fails_when_name_cannot_be_looked_up_at_all():
    with pytest.raises(batou.GetAddressInfoError) as f:
        a = Address("does-not-exist.example.com:1234")
        a.listen
    assert any(x in str(f.value) for x in RESOLVER_ERRORS)

    with pytest.raises(batou.GetAddressInfoError) as f:
        a = Address(
            "does-not-exist.example.com:1234", require_v4=False, require_v6=True
        )
        a.listen_v6
    assert any(x in str(f.value) for x in RESOLVER_ERRORS)

    with pytest.raises(batou.GetAddressInfoError) as f:
        a = Address(
            "does-not-exist.example.com:1234", require_v4=True, require_v6=True
        )
        a.listen
    with pytest.raises(batou.GetAddressInfoError) as f:
        a = Address(
            "does-not-exist.example.com:1234", require_v4=True, require_v6=True
        )
        a.listen_v6
    assert any(x in str(f.value) for x in RESOLVER_ERRORS)


def test_address_optional_when_name_cannot_be_looked_up_at_all():
    a = Address(
        "does-not-exist.example.com:1234",
        require_v4="optional",
        require_v6="optional",
    )
    assert a.listen is None
    assert a.listen_v6 is None


def test_address_required_false_prohibited_to_use():
    a = Address(
        "does-not-exist.example.com:1234", require_v4=True, require_v6=False
    )
    with pytest.raises(batou.IPAddressConfigurationError):
        a.listen_v6

    a = Address(
        "does-not-exist.example.com:1234", require_v4=False, require_v6=True
    )
    with pytest.raises(batou.IPAddressConfigurationError):
        a.listen


def test_address_format_with_port():
    assert str(Address("127.0.0.1:8080").listen) == "127.0.0.1:8080"


@mock.patch("socket.getaddrinfo")
def test_address_should_contain_v6_address_if_available(gai):
    gai.return_value = [(None, None, None, None, ("::1", None, None, None))]
    address = Address("localhost:8080", require_v6=True)
    assert address.listen_v6.host == "::1"


def test_address_prevents_access_to_unconfigured_IPv4_netloc():
    """It raises an exception if IPv4 is not configured but accessed."""
    address = Address("localhost", 22, require_v4=False, require_v6=True)
    with pytest.raises(batou.IPAddressConfigurationError) as err:
        address.listen
    assert (
        "Trying to access address family IPv4 which is not configured for"
        " localhost:22." == str(err.value)
    )


def test_address_prevents_access_to_unconfigured_IPv6_netloc():
    """It raises an exception if IPv6 is not configured but accessed."""
    address = Address("localhost", 22, require_v4=True, require_v6=False)
    with pytest.raises(batou.IPAddressConfigurationError) as err:
        address.listen_v6
    assert (
        "Trying to access address family IPv6 which is not configured for"
        " localhost:22." == str(err.value)
    )


def test_netloc_str_should_brace_ipv6_addresses():
    assert "[::1]:80" == str(NetLoc("::1", 80))


def test_netloc_format_without_port():
    assert str(NetLoc("127.0.0.1")) == "127.0.0.1"


def test_flatten():
    assert [1, 2, 3, 4] == flatten([[1, 2], [3, 4]])


class MultiFileTests(unittest.TestCase):
    def test_write_and_flush_propation(self):
        file1 = StringIO()
        file2 = StringIO()
        multi = MultiFile([file1, file2])
        multi.write("asdf")
        multi.flush()
        self.assertEqual("asdf", file1.getvalue())
        self.assertEqual("asdf", file2.getvalue())


the_fake_lock = threading.Lock()


def fake_lock(lockfile, options):
    # Emulate fcntl's behaviour if we would be two processes.
    success = the_fake_lock.acquire(False)
    if not success:
        raise IOError


class LockfileContextManagerTests(unittest.TestCase):
    def setUp(self):
        self.files = []

    def tearDown(self):
        list(map(os.unlink, self.files))

    def tempfile(self):
        f = tempfile.mktemp()
        self.files.append(f)
        return f

    def test_lock_creates_file_and_writes_and_removes_pid(self):
        lockfile = self.tempfile()
        with locked(lockfile):
            with open(lockfile, "r") as f:
                pid = f.read().strip()
            self.assertEqual(os.getpid(), int(pid))
        with open(lockfile, "r") as f:
            self.assertEqual("", f.read())

    def test_lock_works_with_existing_file(self):
        lockfile = self.tempfile()
        f = open(lockfile, "w")
        f.write("sadf")
        f.close()
        with locked(lockfile):
            with open(lockfile, "r") as f:
                pid = f.read().strip()
            self.assertEqual(os.getpid(), int(pid))
        with open(lockfile, "r") as f:
            self.assertEqual("", f.read())

    @mock.patch("fcntl.lockf", side_effect=fake_lock)
    def test_lock_cant_lock_twice(self, lockf):
        lockfile = self.tempfile()
        with locked(lockfile):
            locked2 = locked(lockfile)
            self.assertRaises(RuntimeError, locked2.__enter__)


class NotifyTests(unittest.TestCase):
    # XXX: What this actually should test is:
    # if notify-send is there, use it. It does not need to exit on all
    # non-darwin machines.
    # @pytest.mark.skipif("sys.platform == 'darwin'")
    # @mock.patch('subprocess.check_call')
    # def test_notify_calls_notify_send(self, call):
    #     notify('foo', 'bar')
    #     call.assert_called_with(['notify-send', 'foo', 'bar'])

    @mock.patch("subprocess.check_call", side_effect=OSError)
    def test_notify_does_not_fail_if_os_call_fails(self, call):
        notify("foo", "bar")


def test_revert_graph_no_edges_is_identical():
    graph = {1: set(), 2: set()}
    assert graph == dict(revert_graph(graph))


def test_revert_graph_one_edge_reverses():
    graph = {1: set([2])}
    assert {2: set([1]), 1: set()} == dict(revert_graph(graph))


def test_topological_sort_simple_chain():
    graph = {1: set([2]), 2: set([3]), 3: set([4])}
    assert [1, 2, 3, 4] == topological_sort(graph)


def test_topological_sort_multiple_paths():
    graph = {1: set([2, 3]), 2: set([3])}
    assert [1, 2, 3] == topological_sort(graph)


def test_topological_sort_raises_on_loop():
    graph = {1: set([2]), 2: set([3]), 3: set([1])}
    with pytest.raises(ValueError):
        topological_sort(graph)


def test_topological_sort_with_single_item():
    graph = {1: set()}
    assert [1] == topological_sort(graph)


def test_graph_remove_leafs():
    graph = {1: [1], 2: []}
    remove_nodes_without_outgoing_edges(graph)
    assert graph == {1: [1]}


class Checksum(unittest.TestCase):
    fixture = os.path.join(
        os.path.dirname(__file__), "fixture", "component", "haproxy.cfg"
    )

    def test_hash_md5(self):
        self.assertEqual(
            "ce0324fa445475e76182c0d114615c7b", hash(self.fixture, "md5")
        )

    def test_hash_sha1(self):
        self.assertEqual(
            "164d8815aa839cca339e38054622b58ca80124a1",
            hash(self.fixture, "sha1"),
        )


@mock.patch("subprocess.Popen")
def test_cmd_joins_list_args(popen):
    popen().communicate.return_value = (b"", b"")
    popen().returncode = 0
    cmd(["cat", "foo", "bar"])
    assert popen.call_args[0] == ("cat foo bar",)


@mock.patch("subprocess.Popen")
def test_cmd_quotes_spacey_args(popen):
    popen().communicate.return_value = (b"", b"")
    popen().returncode = 0
    cmd(["cat", "foo", "bar bz baz"])
    assert popen.call_args[0] == ("cat foo 'bar bz baz'",)
    cmd(["cat", "foo", "bar 'bz baz"])
    assert popen.call_args[0] == (r"cat foo 'bar \'bz baz'",)


@mock.patch("subprocess.Popen")
def test_cmd_ignores_specified_returncodes(popen):
    popen.return_value.returncode = 4
    popen.return_value.communicate.return_value = b"", b""
    with pytest.raises(CmdExecutionError):
        cmd("asdf")
    cmd("asdf", acceptable_returncodes=[0, 4])


@mock.patch("subprocess.Popen")
def test_cmd_returns_process_if_no_communicate(popen):
    process = mock.Mock()
    popen.return_value = process
    p = cmd(["asdf"], communicate=False)
    assert popen.communicate.call_count == 0
    assert p is process


def test_call_with_optional_args():
    def foo():
        return 1

    def bar(x):
        return x

    def baz(**kw):
        return kw["x"]

    def quux(x, y):
        return x

    # The function doesn't expect x, but we're happy to call it.
    assert call_with_optional_args(foo, x=1)
    # The function expect x and it gets passed through
    assert call_with_optional_args(bar, x=4) == 4
    # The function accepts kw args and sees x
    assert call_with_optional_args(baz, x=3) == 3
    # The function accepts x but also expects y and thus breaks
    with pytest.raises(TypeError):
        call_with_optional_args(quux, x=1)


def test_format_duration():
    assert format_duration(1.23124) == "1.23s"
    assert format_duration(61) == "1m1s"
    assert format_duration(None) == "NaN"


def test_export_environment_variables():
    from batou.utils import export_environment_variables

    environ = {
        "a": "1",
        "B": "None",
        "AB": "True",
        "AB1": "'",
        "_A": '"',
        "_1": "${}",
        "_ABBA_": r"\\",
    }
    assert (
        export_environment_variables(environ)
        == """\
export AB1=''"'"''
export AB=True
export B=None
export _1='${}'
export _A='"'
export _ABBA_='\\\\'
export a=1"""
    )

    environ = {"1a": None}
    with pytest.raises(ValueError) as e:
        export_environment_variables(environ)
    assert e.value.args[0] == "`1a` is not a valid shell variable name"

    # This is a real-world example from
    # https://api.wordpress.org/secret-key/1.1/salt/
    environ = {
        "AUTH_KEY": "mc+REHo9-fQ0.GQZDB'bu-/5W&g#+(Lv{+-B`Ai{@nv^UvJ*/|$U6Tzyg>;02p9|Q",
        "SECURE_AUTH_KEY": 'QXVjV-D(~RGX"Rhx(OPr3<j)F}>n4<A .;Ki5hrW}QbPl8%Tv:_$2t+:;}9~bw2SY',
        "LOGGED_IN_KEY": "y{ZO{0N0Ai5Nk|>hk2+!c~WhFJ=-#b1KLu)_B!-Jh2H+rw}R|/pty0ZXGbrq]+,m",
        "NONCE_KEY": "0<>{%+,,u9,Qw,8V*7^je-2* V/%+?c%~KW/Ck[42Vbhf$}3^,kh!*yt3&v,{F-P",
        "AUTH_SALT": "A^hJ)$w7;!5cp$8cFYrDsK.PGN|a-Rn@NR(x@2#-nLZH/LUsnfi,%T3TNmkBF6%s",
        "SECURE_AUTH_SALT": "Vy|F(qC=,Rh61%~A=3>76k5cgpWXZJd/-YiFwx2uacUTWB~F{?Hr::CL%e>AwE-W",
        "LOGGED_IN_SALT": "e|rs/?YVJAhf-40Y:]Sih j||$n-WMyaT|Afp3DWwdmQ(O!FK]w>ZkkpfS*`KlQK",
        "NONCE_SALT": ";FY@_]@c.yiSc+<Q&&YylX8*>iDr7VG5U64/jgoL.$VegW$daH`We _|Q9s~g`|E",
    }
    assert (
        export_environment_variables(environ)
        == """\
export AUTH_KEY='mc+REHo9-fQ0.GQZDB'"'"'bu-/5W&g#+(Lv{+-B`Ai{@nv^UvJ*/|$U6Tzyg>;02p9|Q'
export AUTH_SALT='A^hJ)$w7;!5cp$8cFYrDsK.PGN|a-Rn@NR(x@2#-nLZH/LUsnfi,%T3TNmkBF6%s'
export LOGGED_IN_KEY='y{ZO{0N0Ai5Nk|>hk2+!c~WhFJ=-#b1KLu)_B!-Jh2H+rw}R|/pty0ZXGbrq]+,m'
export LOGGED_IN_SALT='e|rs/?YVJAhf-40Y:]Sih j||$n-WMyaT|Afp3DWwdmQ(O!FK]w>ZkkpfS*`KlQK'
export NONCE_KEY='0<>{%+,,u9,Qw,8V*7^je-2* V/%+?c%~KW/Ck[42Vbhf$}3^,kh!*yt3&v,{F-P'
export NONCE_SALT=';FY@_]@c.yiSc+<Q&&YylX8*>iDr7VG5U64/jgoL.$VegW$daH`We _|Q9s~g`|E'
export SECURE_AUTH_KEY='QXVjV-D(~RGX"Rhx(OPr3<j)F}>n4<A .;Ki5hrW}QbPl8%Tv:_$2t+:;}9~bw2SY'
export SECURE_AUTH_SALT='Vy|F(qC=,Rh61%~A=3>76k5cgpWXZJd/-YiFwx2uacUTWB~F{?Hr::CL%e>AwE-W'"""
    )



import mock
import pytest

from ..host import RPCWrapper


def test_rpcwrapper_returns_result():
    host = mock.Mock()
    host.channel.receive.return_value = ("batou-result", mock.sentinel.result)
    rpc = RPCWrapper(host)
    assert mock.sentinel.result == rpc.foo()


def test_rpcwrapper_error_contains_hostname():
    host = mock.Mock()
    host.fqdn = "foo.example.com"
    host.channel.receive.return_value = ("batou-error", None)
    rpc = RPCWrapper(host)
    with pytest.raises(RuntimeError) as e:
        rpc.foo()
    assert ("foo.example.com: Remote exception encountered.",) == e.value.args



import os.path

from batou import UpdateNeeded
from batou.component import Component
from batou.lib.file import Directory


class Checkout(Component):

    namevar = "url"
    target = "."
    revision = None

    def configure(self):
        self += Directory(self.target)

    def verify(self):
        with self.chdir(self.target):
            if not os.path.exists(".svn"):
                raise UpdateNeeded()
            stdout, stderr = self.cmd("svn info | grep Revision:")
            current_revision = stdout.replace("Revision:", "").strip()
            if current_revision != self.revision:
                raise UpdateNeeded()

    def update(self):
        with self.chdir(self.target):
            if not os.path.exists(".svn"):
                self.cmd(
                    self.expand(
                        "svn co {{component.url}} . -r {{component.revision}}"
                    )
                )
            else:
                self.cmd(self.expand("svn revert -R ."))
                self.cmd(self.expand("svn switch {{component.url}}"))
                self.cmd(self.expand("svn up -r {{component.revision}}"))


Subversion = Checkout  # BBB



import os.path
import unittest

import mock
import pytest

from ..download import Download


class DownloadTest(unittest.TestCase):
    def test_verify_should_pass_checkum_function_to_hash(self):
        import batou

        component = Download("url", checksum="foobar:1234")
        component.configure()
        with mock.patch("os.path.exists") as exists, mock.patch(
            "batou.utils.hash"
        ) as buh:
            exists.return_value = True
            try:
                component.verify()
            except batou.UpdateNeeded:
                pass
        buh.assert_called_with(mock.ANY, "foobar")

    def test_configure_should_raise_valueerror_if_no_checksum_given(self):
        download = Download("url")
        with self.assertRaises(ValueError) as exc:
            download.configure()
        self.assertEqual("No checksum given.", exc.exception.args[0])

    def test_update_should_raise_AssertionError_on_checksum_mismatch(self):
        download = Download("url", checksum="foobar:1234")
        download.configure()
        with mock.patch(
            "batou.lib.download.urlretrieve"
        ) as retrieve, mock.patch("batou.utils.hash") as buh, self.assertRaises(
            AssertionError
        ) as err:
            retrieve.return_value = "url", []
            buh.return_value = "4321"
            download.update()
        self.assertEqual(
            "Checksum mismatch!\nexpected: 1234\ngot: 4321", str(err.exception)
        )

    def test_breadcrumbs_should_not_show_password(self):
        component = Download(
            "http://me:mypw@myhost/file.name", checksum="foobar:1234"
        )
        assert (
            "http://me:*****@myhost/file.name"
            == component.namevar_for_breadcrumb
        )


@pytest.mark.slow
def test_downloads_file(root):
    root.component += Download(
        "https://downloads.fcio.net/batou-tests/test100k",
        checksum="sha256:4e6424e0151f79bda49bd3769052b29"
        "12793c258d066689865b69abe4e696452",
    )
    root.component.deploy()
    assert os.path.isfile(
        os.path.join(root.environment.workdir_base, "mycomponent/test100k")
    )



import os

import pytest

from batou.tests.ellipsis import Ellipsis


def test_main_with_errors(capsys):
    os.chdir("examples/errors")

    from batou.deploy import main

    with pytest.raises(SystemExit) as r:
        main(
            environment="errors",
            platform=None,
            timeout=None,
            dirty=False,
            consistency_only=False,
            predict_only=False,
            check_and_predict_local=False,
            jobs=None,
            provision_rebuild=False,
        )

    assert r.value.code == 1

    out, err = capsys.readouterr()
    assert err == ""
    # save the output to a file to compare it with the expected output
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `errors`...
main: Verifying repository ...
main: Loading secrets ...

ERROR: Failed loading component file
           File: .../examples/errors/components/component5/component.py
      Exception: invalid syntax (component.py, line 1)
Traceback (simplified, most recent call last):
<no-non-remote-internal-traceback-lines-found>

ERROR: Failed loading component file
           File: .../examples/errors/components/component6/component.py
      Exception: No module named 'asdf'
Traceback (simplified, most recent call last):
  File ".../examples/errors/components/component6/component.py", line 1, in <module>
    import asdf  # noqa: F401 import unused
...
ERROR: Missing component
      Component: missingcomponent

ERROR: Superfluous section in environment configuration
        Section: superfluoussection

ERROR: Override section for unknown component found
      Component: nonexisting-component-section

ERROR: Attribute override found both in environment and secrets
      Component: component1
      Attribute: my_attribute

ERROR: Secrets section for unknown component found
      Component: another-nonexisting-component-section
======================= DEPLOYMENT FAILED (during load) ========================
"""
    )  # noqa: E501 line too long


def test_main_fails_if_no_host_in_environment(capsys):
    os.chdir("examples/errorsnohost")

    from batou.deploy import main

    with pytest.raises(SystemExit) as r:
        main(
            environment="errorsnohost",
            platform=None,
            timeout=None,
            dirty=False,
            consistency_only=False,
            predict_only=False,
            check_and_predict_local=False,
            jobs=None,
            provision_rebuild=False,
        )

    assert r.value.code == 1

    out, err = capsys.readouterr()
    assert err == ""
    assert out == Ellipsis(
        """\
batou/2... (cpython 3...)
================================== Preparing ===================================
main: Loading environment `errorsnohost`...
main: Verifying repository ...
main: Loading secrets ...
================== Connecting hosts and configuring model ... ==================

ERROR: No host found in environment.
====================== DEPLOYMENT FAILED (during connect) ======================
"""
    )  # noqa: E501 line too long



from pathlib import Path


def migrate(output):
    """Rename secret files to match new naming scheme."""
    output(
        "Secrets may be encrypted with age or gpg now",
        """

The two secret providers `age` and `gpg` signify their encryption method in the
file name. The new naming scheme is:
secrets.cfg -> secrets.cfg.gpg
secret-foo -> secret-foo.gpg

Since previously all secrets were encrypted with gpg, we rename all files to
match the new naming scheme. This is done automatically for you, but you will
need to commit those changes manually.

        """,
        "automatic",
    )
    for name in _get_environment_names():
        _migrate_environment(name)


def _get_environment_names():
    return [x for x in (Path.cwd() / "environments").iterdir() if x.is_dir()]


def _migrate_environment(name: str) -> None:
    cwd = Path.cwd()
    environment_dir = cwd / "environments" / name
    secrets_cfg = environment_dir / f"secrets.cfg"
    if secrets_cfg.exists():
        secrets_cfg.rename(str(secrets_cfg) + ".gpg")
        for secret_file in environment_dir.iterdir():
            if secret_file.name.startswith(
                "secret-"
            ) and not secret_file.name.endswith(".gpg"):
                secret_file.rename(str(secret_file) + ".gpg")
